import numpy as np
import random as rd


def iterative_filter(
    func: callable,
    variables: dict,
    nsamples: int,
    thresholds: list,
    conditions: dict,
    folder: str,
    max_repetitions: int = 3,
    index: list = None,
    controls: list = None,
):
    """Performs the iterative filtering calibration method on a given function with uncertain inputs.

    args:

        func (callable): the function being analysed and used to get the outputs.

        variables (dict): a dictionary that contains the names and bounds of all variables within the function.

        nsamples (int): the number of samples taken for each variable.

        thresholds (list): the set of threshold filter values that the conditions will need to meet.

        conditions (dict): a dictionary of 'correct' values that the outputs of interest will need to meet.

        folder (str): where all outputs of this function will be stored in.

        max_repetitions (int): how many repetitions are allowed at a singular threshold value before stopping the process.

        index (list): a list of indexes referring to the list of variables that are being filtered.

        controls (list): a list of indexes referring to the variables whose values are unique to their condition and must be generated separately.
    """

    if (
        index == None
    ):  # If i is None, that means that no variables are selected for filtering which does not make sense and so returns an error
        raise ValueError(
            "No variables selected for filtering, use the index to list variables to be filtered."
        )

    for s in index:
        with open(f"./{folder}/bounds_all_min.txt", "a") as a:
            a.write(str(variables[conditions["names"][0]][s][0]) + ",")
        with open(f"./{folder}/bounds_all_max.txt", "a") as a:
            a.write(str(variables[conditions["names"][0]][s][1]) + ",")
    with open(f"./{folder}/bounds_all_min.txt", "a") as a:
        a.write(f"str(0)" + "," + f">{thresholds[0]}" + "\n")
    with open(f"./{folder}/bounds_all_max.txt", "a") as a:
        a.write(f"str(0)" + "," + f">{thresholds[0]}" + "\n")

    Iteration = 1
    max_Iteration = len(thresholds[1:])
    repetitions = 0
    total_repetitions = 0
    nconditions = len(conditions["names"])

    while Iteration <= max_Iteration:
        print(Iteration)  # optional, provides visual counter of progress
        X_input = np.empty([nsamples * nconditions, len(variables["names"])])
        for j in range(0, nsamples * nconditions, 2):
            for k in range(nconditions):
                for l in range(len(variables["names"])):
                    if l not in controls and k > 0:
                        X_input[j + k, l] = X_input[j, l]
                    elif (
                        variables["number_type"][l] == "int"
                    ):  # Used for NASA UQ Example to get an integer for the seed.
                        X_input[j + k, l] = rd.randint(
                            variables[conditions["names"][k]][l][0],
                            variables[conditions["names"][k]][l][1],
                        )
                    else:
                        X_input[j + k, l] = rd.uniform(
                            variables[conditions["names"][k]][l][0],
                            variables[conditions["names"][k]][l][1],
                        )

        # for j in range(nconditions): #Storing each condition's inputs separately (optional)
        #     X_input_set = X_input[j::nconditions]
        #     input_file_path = f'./{folder}/input{variables['names'][j]}_Level{Iteration}_{repetitions}.txt'
        #     np.savetxt(input_file_path, X_input_set, delimiter=',')

        input_file_path = f"./{folder}/input_Level{Iteration}_{repetitions}.txt"
        np.savetxt(
            input_file_path, X_input, delimiter=","
        )  # stores the full input matrix in one file (required for NASA UQ example)

        Y_out = func(
            input_file_path, nsamples, nconditions
        )  # Must be edited as per user input requirements, output must be 'Y_out'.

        if (
            repetitions == 0
        ):  # When a new Iteration is reached, resets the accepted sample matrix and count.
            filtered_samples = np.empty([nsamples, 9])
            count = 0
        for i in range(nsamples):  # Looping through all samples
            test_threshold = thresholds[Iteration]
            test_Level = Iteration
            point_added = False
            while (
                point_added is not True
            ):  # While loop finds which Iteration the sample belongs to.
                threshold_set = np.empty([len(conditions["values"]), nconditions])
                for j in range(len(conditions["values"])):
                    threshold_set[j] = (
                        abs(Y_out[j, i, :] - conditions["values"][j]) <= test_threshold
                    )

                if threshold_set[:, :].all() == 1:
                    if (
                        test_Level == Iteration
                    ):  # If sample is accepted at the same Iteration that it starts, it is added to the matrix for updating the Xe bounds
                        filtered_samples[count] = X_input[nconditions * i]
                        count += 1
                    point_added = True
                    for s in index:
                        with open(f"./{folder}/samples.csv", "a") as a:
                            a.write(str(X_input[nconditions * i, s]) + ",")
                        with open(f"./{folder}/Iteration{test_Level}.csv", "a") as a:
                            a.write(str(X_input[nconditions * i, s]) + ",")
                        with open(f"./{folder}/samples.csv", "a") as a:
                            a.write(f"{test_Level}" + "," + "\n")
                        with open(f"./{folder}/Iteration{test_Level}.csv", "a") as a:
                            a.write(f"{test_Level}" + "," + "\n")
                else:
                    test_Level = test_Level - 1
                    test_threshold = thresholds[
                        test_Level
                    ]  # Moving the radius down a Iteration (can be changed out to an array of radii instead)
                    if (
                        test_Level <= 0
                    ):  # If sample is not within the start radius for all controls, the while loop is broken.
                        for s in index:
                            with open(f"./{folder}/samples.csv", "a") as a:
                                a.write(str(X_input[nconditions * i, s]) + ",")
                            with open(
                                f"./{folder}/Iteration{test_Level}.csv", "a"
                            ) as a:
                                a.write(str(X_input[nconditions * i, s]) + ",")
                            with open(f"./{folder}/samples.csv", "a") as a:
                                a.write(f"{test_Level}" + "," + "\n")
                            with open(
                                f"./{folder}/Iteration{test_Level}.csv", "a"
                            ) as a:
                                a.write(f"{test_Level}" + "," + "\n")
                        break

        print(
            count
        )  # Optional, allows for a quick check of how many samples were 'accepted' at a given iteration while process is ongoing.
        if (
            count > 1
        ):  # If more than one sample is accepted at the current Iteration, new bounds are computed and Iteration goes up.
            filtered_samples.resize(count, len(variables["names"]))

            for i in range(nconditions):
                for s in index:
                    variables[conditions["names"][i]][s] = np.min(
                        filtered_samples[:, s]
                    ), np.max(filtered_samples[:, s])
                    if i == 0:
                        print(
                            f"New Bounds for {variables['names'][s]}: {np.min(filtered_samples[:,s]), np.max(filtered_samples[:,s])}"
                        )
            for s in index:
                with open(f"./{folder}/bounds_all_min.txt", "a") as a:
                    a.write(str(variables[conditions["names"][i]][s][0]) + ", ")
                with open(f"./{folder}/bounds_all_max.txt", "a") as a:
                    a.write(str(variables[conditions["names"][i]][s][1]) + ", ")
            with open(f"./{folder}/bounds_all_min.txt", "a") as a:
                a.write(
                    f" str({Iteration})"
                    + ", "
                    + f"{thresholds[Iteration]}"
                    + ", "
                    + "\n"
                )
            with open(f"./{folder}/bounds_all_max.txt", "a") as a:
                a.write(
                    f"str({Iteration})"
                    + ", "
                    + f"{thresholds[Iteration]}"
                    + ", "
                    + "\n"
                )

            total_repetitions += (
                repetitions  # Add repetitions from current Iteration to total
            )
            repetitions = 0  # Reset repetitions when new bounds are found.
            Iteration += (
                1  # Increase Iteration (and by extension, decrease radius being tested)
            )
            if (
                Iteration > max_Iteration
            ):  # If max_Iteration is reached, no more radii are available so while loop is broken.
                print(
                    f"All control points intersect within the radius {thresholds[max_Iteration]} with {max_Iteration} levels of reduction and {max_Iteration+total_repetitions} iterations of {nsamples} samples."
                )
                break

        else:  # Otherwise the process gets repeated and the number of repetitions at the current Iteration is increased.
            if (
                repetitions >= max_repetitions
            ):  # If Iteration is repeated max_repetitions amount of times, while loop is broken.
                total_repetitions += repetitions
                print(
                    f"All control points lack viable points below radius: {thresholds[Iteration]} at Iteration {Iteration} with {Iteration+total_repetitions} iterations of {nsamples} samples."
                )
                break
            repetitions += 1

    return



import numpy as numpy
from numpy.typing import NDArray
from .scenario import minimal_enclosing_hyperbox, is_inside_box, epsLU
from .fuzzy import boxes_to_fuzzy_projection
from scipy import stats

"""  Hints on notation used in this module:

### Forward problem ###
x: data set, each sample is a (nxd) vector.
d: dimension of each vector.
n: number of samples.

----
a: list (of length l) of subindices corresponding to the support vectors.
b: list (of length l) of enclosing sets.
l: number of levels.
p: violation upper probabilities, or membership values.
f: array (lxdx2) of projections of enclosing sets, for marginal plots.

### Backward problem ###
y: data set, each sample is a (nxd) vector.
d: dimension of each vector in the data set, or dimension of the output space.
n: number of samples.
`f`: a model or function such that `y=f(x)`.
`ux`: array (mxd_) of coverage samples.
m: number of coverage samples.
d_: dimension of input space.
uy: array (mxd) of evaluated coverage samples, such that `uy=f(ux)`.

----
a: list (of length l) of subindices corresponding to the support vectors.
b: list (of length l) of enclosing sets.
l: number of levels.
c: list (of length l) of subindices of coverage samples belonging to each level.
p: list (of length l) of violation upper probabilities, or membership values.
fx: array (lxd_x2) of projections of enclosing sets input space.
fy: array (lxdx2) of projections of enclosing sets output space.
"""


def data_peeling_algorithm(
    X: NDArray, tol: float = 1e-4
) -> tuple[list, list]:  # should also input the shape, rectangle, circle, etc.
    """Data peeling algorithm for constructing a sequence of nested enclosing sets.

    args:
        X (NDArray): data set of iid observations
        tol (float): a tolerance parameter determining the minimal size of an allowed enclosing box

    returns:
        tuple[list, list]: a tuple containing:
            - sequence_of_indices (list):
                a list (number of levels long) of sets (or list) of indices of heterogeneous size
            - sequence_of_boxes (list):
                a list (number of levels long) of (dx2) boxes


    .. figure:: /_static/peeling_illus.png
        :alt: ks_bounds
        :align: center
        :width: 90%

        Illustration of the data peeling algorithm.
    """
    H = make_hashtable(X)  # turns dataset in to hashtable
    n, d = X.shape
    sequence_of_indices, sequence_of_boxes = [], []
    index_num = 0  # number of support vectors per level
    j = 0  # number of levels
    while index_num < n:
        if j == 0:
            active_indices, box = minimal_enclosing_hyperbox(
                X
            )  # solve the optimization
            active_mask = index_to_mask(active_indices, n=n)
            nonactive_mask = active_mask == False
            X_strip = X[active_mask, :].copy()
            collect_indices = []
            for xstrip in X_strip:
                collect_indices.append(H[tuple(xstrip)])
                index_num += 1
            X_new = X[
                nonactive_mask, :
            ].copy()  # strip active scenarios from the original dataset
        else:
            active_indices, box = minimal_enclosing_hyperbox(X_new)
            active_mask = index_to_mask(active_indices, n=X_new.shape[0])
            nonactive_mask = active_mask == False
            X_strip = X_new[active_mask, :].copy()
            X_new = X_new[nonactive_mask, :].copy()
            collect_indices = []
            for xstrip in X_strip:
                collect_indices.append(H[tuple(xstrip)])
                index_num += 1
        sequence_of_indices.append(collect_indices)
        sequence_of_boxes.append(box)
        j += 1
    # return sequence_of_indices, sequence_of_boxes
    a, b = sanity_check(
        sequence_of_indices, sequence_of_boxes, tol=tol
    )  # this checks that the algorithm terminates as it is supposed to.
    return a, b


def index_to_mask(x: list, n=100) -> numpy.ndarray:
    """Converts the list of indices x into a (nx1) boolean mask, true at those particular indices

    Args:
        x (list): list of indices
        n (int, optional): size of the mask. Defaults to 100.

    Returns:
        numpy.ndarray: boolean mask
    """
    nrange = numpy.arange(n)
    boole = numpy.zeros(nrange.shape, dtype=bool)
    for xi in x:
        boole += nrange == xi
    return boole


def make_hashtable(x):
    h = {}
    for i, xi in enumerate(x):
        h[tuple(xi)] = i
    return h


def sanity_check(
    a, b, tol=1e-4
):  # checks that the smallest set has enough observation to make a decsion.
    b_last = b[-1]  # last enclosing set in the sequence
    for interval in b_last:
        if width(interval) < tol:
            return a[:-1], b[:-1]  # returns sequences stripped of the last element
    return a, b


def data_peeling_backward(
    uy: NDArray, y: NDArray = None, boxes: list = None, tol=1e-4
) -> tuple[list, list, list]:
    """
    Args:
        uy (NDArray): (mxd) array of coverage samples output space.
        boxes (list): sequence of boxes, each box is a (dx2) array. Also iterable of interval objects.

    Returns:
        tuple[list, list, list]: a tuple containing:
            - a: list (number of levels long) of sets (or list) of indices (heterogeneous size).
            - b: list (number of levels long) of (dx2) boxes (array-like).
            - c: list (number of levels long) of indices (input space) conatained in each level.
    Note:
        There are two cases where the peeling algorithm must raise an exception, and they are both linked to the termination of the algorithm.
            - When the last enclosing set has less samples than the minimum number of support scenarios to determine that set.
            - When there are enough samples to determine the set but some of them are too close to eachother (even for just one dimension).

    While two may be linked to the problem of degeneracy, in this context, it may be best suited to refer to this case as a coverage problem.
    """
    a = None  # initialise in case peeling forward is not called
    if (boxes is None) & (y is not None):
        a, b = data_peeling_algorithm(y, tol=tol)
    contained_in_each_level = []
    for box in b:
        contained_in_each_level.append(is_inside_box(uy, box))  # from bottom to top
    return (
        a,
        b,
        contained_in_each_level,
    )  # indices of samples contained in each level from bottom to top


def extract_kn(a):
    kk = 0
    k_cum = []
    for subi in a:
        kk += len(subi)
        k_cum.append(kk)
    return k_cum, kk


def peeling_to_structure(
    a: list, b: list, kind: str = "scenario", beta: float = 0.01
) -> tuple[NDArray, list]:
    """Peeling to structure.

    Args:
        a (list):  (number of levels long) of sets (or list) of indices of heterogeneous size
        b (list):  (number of levels long) of (dx2) boxes (array-like)

    Returns:
        f: (lxdx2) array containing the projections of a joint fuzzy number.
        p: list[float] of length l

    Note:
        Note b may not be boxes, but spheres, ellipses, etc.

    .. tip::
        keep b for piping: peeling_to_structure(data_peeling_algorithm(x))
    """
    k_cum, n = extract_kn(a)
    ALPHA = []
    ALPHA_append = ALPHA.append
    for k in k_cum:
        if kind == "scenario":
            alphaL, alphaU = epsLU(k, n, beta)
        elif kind == "c-box":
            pass
        elif kind == "clopper-pearson":
            pass
        elif kind == "percentage_sets":
            pass
        elif kind == "uniform":
            pass
        ALPHA_append(alphaU)  # each line may take a while if n is large
    f, p = boxes_to_fuzzy_projection(b, p=ALPHA)
    return f, p


def uniform(lo, hi, N=100):
    lo = numpy.asarray(lo, dtype=float)
    hi = numpy.asarray(hi, dtype=float)
    if lo.size == 1:
        d = 1
    else:
        d = lo.shape[0]  # if d == 0: stats.uniform(loc=lo, scale=hi-lo).rvs((N,))
    return stats.uniform(loc=lo, scale=hi - lo).rvs((N, d))


def samples_to_structure(ux: NDArray, c: list):
    """
    Args:
        ux (NDArray): (mxd_) array.
        c (list): (number of levels long) subset of indices (input space) conatained in each level.

    Returns:
        fx: (lxd_x2) array containing the projections of a joint fuzzy number in the input space.
    """

    pass


def width(x: NDArray):
    """
    Args:
        x: An interval or interval iterable, i.e. an (dx2) array

    Returns:
        w: the width of the intervals
    """
    x = numpy.asarray(x, dtype=float)
    if len(x.shape) == 1:
        return x[1] - x[0]
    else:
        return x[:, 1] - x[:, 0]



from numpy.typing import NDArray
import matplotlib.pyplot as plt
from scipy.spatial import ConvexHull
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import numpy as np
from scipy.spatial import HalfspaceIntersection, ConvexHull
from scipy.optimize import linprog
from functools import partial


class EpistemicFilter:
    """The EpistemicFilter method to reduce the epistemic uncertainty space based on discrepancy scores.

    args:
        xe_samples (NDArray): Proposed Samples of epistemic parameters, shape (ne, n_dimensions).
            Typically samples from a bounded set of some epistemic parameters.

        discrepancy_scores (NDArray, optional): Discrepancy scores between the model simulations and the observation.
            Associated with each xe sample, shape (ne,). Defaults to None.

        sets_of_discrepancy (list, optional): List of sets of discrepancy scores for multiple datasets.
            Each element should be an NDArray of shape (ne,). Defaults to None.

    tip:
        For performance functions that output multiple responses, some aggregation of discrepancy scores may be used.
        Depending on the number of observation, either a single set of discrepancy scores or multiple sets can be provided.

    .. figure:: /_static/convex_hull.png
        :alt: convex hull with bounds
        :align: center
        :width: 50%

        Convex hull with bounds illustration.
    """

    def __init__(
        self,
        xe_samples: NDArray,
        discrepancy_scores: NDArray = None,
        sets_of_discrepancy: list = None,
    ):

        self.xe_samples = xe_samples
        self.discrepancy_scores = discrepancy_scores
        self.sets_of_discrepancy = (
            sets_of_discrepancy if sets_of_discrepancy is not None else None
        )

    def filter(self, threshold: float | list):
        """Filter the epistemic samples based on a discrepancy threshold.

        args:
            threshold (float | list): The discrepancy threshold(s) for filtering data points.

        returns:
            tuple:
                - the filtered xe samples;
                - hull;
                - lower bounds;
                - upper bounds of the bounding box, or (None, None) if unsuccessful.
        """
        if isinstance(threshold, list):
            return self.iterative_filter(thresholds=threshold)
        else:
            return filter_by_discrepancy(
                self.xe_samples, self.discrepancy_scores, threshold
            )

    def iterative_filter(
        self,
        thresholds: list,
    ) -> list:
        """Iteratively filter the epistemic samples based on multiple thresholds.

        args:
            thresholds (list): List of discrepancy thresholds for filtering data points.

            simulation_model (callable, optional): A function that takes xe_samples as input and returns discrepancy scores.
                Required if re_sample is True. Defaults to None.

        note:
            thresholds are assumed to be sorted in ascending order.
        """

        f = partial(
            filter_by_discrepancy,
            xe_samples=self.xe_samples,
            discrepancy_scores=self.discrepancy_scores,
        )
        return [f(t) for t in thresholds.sort()]

    # TODO develope this function which natively intergrates the simulation model and resampling
    def iterative_filter_w_simulation_model(
        self, thresholds: list, re_sample: bool = False, simulation_model=None
    ):
        """Iteratively filter the epistemic samples based on multiple thresholds.

        args:
            thresholds (list): List of discrepancy thresholds for filtering data points.

            re_sample (bool, optional): If True, re-evaluate discrepancy scores at each iteration using the simulation_model.
                Defaults to False.

            simulation_model (callable, optional): A function that takes xe_samples as input and returns discrepancy scores.
                Required if re_sample is True. Defaults to None.
        note:
            if re_sample is True, the simulation_model should be provided to re-evaluate discrepancy scores.
        """

        pass

    def filter_on_sets(self, plot_hulls: bool = True):
        """Filter the epistemic samples based on multiple sets of discrepancy scores.

        args:
            plot_hulls (bool, optional): If True, plots the convex hulls of the first five sets. Defaults to True.

        returns:
            tuple:
                - lower bounds;
                - upper bounds of the intersected bounding box, or (None, None) if unsuccessful

        note:
            `self.sets_of_discrepancy` must exist.
        """

        xe_list = []
        hull_list = []
        boxes = []

        for i in range(len(self.sets_of_discrepancy)):
            ef = EpistemicFilter(
                xe_samples=self.xe_samples,
                discrepancy_scores=self.sets_of_discrepancy[i],
            )
            filtered_xe, hull, low_bound, upper_bound = ef.filter(threshold=10)
            xe_list.append(filtered_xe)
            hull_list.append(hull)
            boxes.append((hull.min_bound, hull.max_bound))

        if plot_hulls:
            plot_multiple_convex_hulls(xe_list[:5], hull_list[:5])

        # Compute intersection of bounding boxes
        box_int_lower = np.maximum(boxes[0][0], boxes[1][0])
        box_int_upper = np.minimum(boxes[0][1], boxes[1][1])

        return box_int_lower, box_int_upper

    def plot_hull_with_bounds(
        self,
        filtered_xe,
        hull=None,
        ax=None,
        show=True,
        x_title=None,
        y_title=None,
        hull_alpha=0.25,
    ):
        """Plot the convex hull and bounding box of the epistemic samples.

        args:
            hull (ConvexHull, optional): Precomputed convex hull. If None, it is computed.

            ax (matplotlib Axes, optional): Existing axes to draw on. If None, a new figure/axes is created.

            show (bool, optional): If True, calls plt.show() at the end. Defaults to True.

            x_title (str, optional): Label for the x-axis. Defaults to None.

            y_title (str, optional): Label for the y-axis. Defaults to None.

            hull_alpha (float, optional): Transparency of the hull surface. Defaults to 0.25.

        returns:
            ax (matplotlib Axes): The axes containing the plot.
        """
        return plot_convex_hull_with_bounds(
            filtered_xe,
            hull=hull,
            ax=ax,
            show=show,
            x_title=x_title,
            y_title=y_title,
            hull_alpha=hull_alpha,
        )


# problem agnostic
def filter_by_discrepancy(xe_samples, discrepancy_scores, threshold=0.1) -> tuple:
    """Computes the intersection of convex hull bounding boxes based on a discrepancy threshold.

    args:
        threshold (float): The scalar discrepancy threshold for filtering data points.

    returns:
        tuple: the hull, and Lower and upper bounds of the intersected bounding box, or (None, None) if unsuccessful.

    note:
        Assume the threshold is a scalar value.
    """
    # Get the absolute path of the directory containing this script

    filtered_xe = xe_samples[discrepancy_scores < threshold]

    if filtered_xe.size == 0:
        raise ValueError(f"No data points remain after filtering.")

    hull = ConvexHull(filtered_xe)

    # return hull, hull.min_bound, hull.max_bound

    # for multiple thresholds/datasets
    # convex_hulls = []
    # boxes = []

    # convex_hulls.append(hull)
    # boxes.append((hull.min_bound, hull.max_bound))

    # if len(boxes) < 2:
    #     print("Error: Not enough valid bounding boxes to compute intersection.")
    #     return None, None

    # # Compute intersection of bounding boxes
    # box_int_lower = np.maximum(boxes[0][0], boxes[1][0])
    # box_int_upper = np.minimum(boxes[0][1], boxes[1][1])

    # return box_int_lower, box_int_upper

    return filtered_xe, hull, hull.min_bound, hull.max_bound


def plot_convex_hull_with_bounds(
    filtered_xe: NDArray,
    hull=None,
    ax=None,
    show=True,
    x_title=None,
    y_title=None,
    hull_alpha=0.25,
):
    """Plot points, their convex hull, and the axis-aligned bounding box.

    args:
    filtered_xe (NDArray): array-like, shape (n_samples, n_dims)
        Input points. Supports 2D or 3D.

    hull : scipy.spatial.ConvexHull, optional
        Precomputed convex hull. If None, it is computed.

    ax : matplotlib Axes or 3D Axes, optional
        Existing axes to draw on. If None, a new figure/axes is created.

    show : bool, default True
        If True, calls plt.show() at the end.

    hull_alpha : float, default 0.25
        Transparency of the hull surface.

    Returns:
        fig : matplotlib.figure.Figure
        ax : matplotlib.axes.Axes or mplot3d.Axes3D
        hull : scipy.spatial.ConvexHull
        min_bound : ndarray
        max_bound : ndarray
    """
    xe = np.asarray(filtered_xe)
    if xe.ndim != 2:
        raise ValueError("xe must be a 2D array of shape (n_samples, n_dims).")

    n_dims = xe.shape[1]
    if n_dims not in (2, 3):
        raise ValueError("Only 2D or 3D data can be plotted.")

    # Compute hull if needed
    if hull is None:
        hull = ConvexHull(xe)

    # SciPy provides these; fall back to data min/max if not present
    min_bound = getattr(hull, "min_bound", xe.min(axis=0))
    max_bound = getattr(hull, "max_bound", xe.max(axis=0))

    # Set up figure/axes
    if ax is None:
        if n_dims == 2:
            fig, ax = plt.subplots()
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
    else:
        fig = ax.figure

    # -------------------- 2D CASE --------------------
    if n_dims == 2:
        # Scatter points
        ax.scatter(xe[:, 0], xe[:, 1], s=20, color="blue")

        # Convex hull polygon
        verts = xe[hull.vertices]
        verts_closed = np.vstack([verts, verts[0]])  # close polygon

        ax.fill(
            verts_closed[:, 0],
            verts_closed[:, 1],
            edgecolor="r",
            facecolor="r",
            alpha=hull_alpha,
            linewidth=2,
        )

        # Bounding rectangle
        xmin, ymin = min_bound
        xmax, ymax = max_bound
        rect_x = [xmin, xmax, xmax, xmin, xmin]
        rect_y = [ymin, ymin, ymax, ymax, ymin]

        ax.plot(rect_x, rect_y, "--", linewidth=2, color="g")

        # ax.set_aspect("equal", "box")

    # -------------------- 3D CASE --------------------
    else:
        # Scatter points
        ax.scatter(xe[:, 0], xe[:, 1], xe[:, 2], s=20, color="C0", depthshade=True)

        # Convex hull faces
        faces = [xe[simplex] for simplex in hull.simplices]
        poly = Poly3DCollection(faces, alpha=hull_alpha)
        poly.set_edgecolor("r")
        poly.set_facecolor("r")
        ax.add_collection3d(poly)

        # Bounding box (axis-aligned)
        xmin, ymin, zmin = min_bound
        xmax, ymax, zmax = max_bound

        corners = np.array(
            [
                [xmin, ymin, zmin],
                [xmax, ymin, zmin],
                [xmax, ymax, zmin],
                [xmin, ymax, zmin],
                [xmin, ymin, zmax],
                [xmax, ymin, zmax],
                [xmax, ymax, zmax],
                [xmin, ymax, zmax],
            ]
        )

        edges = [
            (0, 1),
            (1, 2),
            (2, 3),
            (3, 0),  # bottom square
            (4, 5),
            (5, 6),
            (6, 7),
            (7, 4),  # top square
            (0, 4),
            (1, 5),
            (2, 6),
            (3, 7),  # vertical edges
        ]

        for i, j in edges:
            xs, ys, zs = zip(corners[i], corners[j])
            ax.plot(xs, ys, zs, "--", linewidth=2, color="g")

        # Nice aspect ratio
        ax.set_box_aspect(max_bound - min_bound)

    # Labels
    if x_title is not None:
        ax.set_xlabel(x_title)
    else:
        ax.set_xlabel("x")

    if y_title is not None:
        ax.set_ylabel(y_title)
    else:
        ax.set_ylabel("y")

    if n_dims == 3:
        ax.set_zlabel("z")

    if show:
        plt.show()

    return ax


def intersect_convex_hulls(hulls):
    """Compute intersection of several ConvexHull polytopes.
    Works in 2D or 3D (and in principle higher).

    args:
        hulls : list of scipy.spatial.ConvexHull

    returns:
        vertices : (m, d) ndarray
            Vertices of the intersection polytope (may be empty).
        hull_int : ConvexHull or None
            ConvexHull of the intersection vertices, or None if empty.
    """
    if not hulls:
        raise ValueError("Need at least one hull.")

    dim = hulls[0].points.shape[1]

    # 1. Collect all halfspaces: a^T x + c <= 0
    halfspaces = np.vstack([h.equations for h in hulls])

    # 2. Try to find a feasible interior point by linear programming
    A = halfspaces[:, :-1]
    c = halfspaces[:, -1]

    # HalfspaceIntersection expects a^T x + c <= 0.
    # We solve A x + c <= 0  <=>  A x <= -c
    res = linprog(
        np.zeros(dim),
        A_ub=A,
        b_ub=-c,
        method="highs",
    )

    if not res.success:
        # Intersection is empty (or numerical issues)
        return np.empty((0, dim)), None

    interior_point = res.x

    # 3. Build the halfspace intersection
    hs_int = HalfspaceIntersection(halfspaces, interior_point)
    vertices = hs_int.intersections

    if len(vertices) == 0:
        return vertices, None

    hull_int = ConvexHull(vertices)
    return vertices, hull_int


def plot_multiple_convex_hulls(
    xe_list,
    hulls=None,
    show_bounds=True,
    hull_alpha=0.25,
    ax=None,
    show=True,
):
    """Plot multiple 2D point sets with their convex hulls.

    args:
        xe_list : list of ndarray
            Each array must be shape (n_i, 2)
        hulls : list of ConvexHull or None
            If None, hulls are computed automatically.
        show_bounds : bool
            Whether to draw dashed bounding rectangles.
        hull_alpha : float
            Transparency of hull fill.
        ax : matplotlib Axes
            Optional existing axes.
        show : bool
            Whether to call plt.show()

    Returns:
        fig, ax, hulls
    """

    # Convert inputs and validate dimensions
    xe_list = [np.asarray(xe) for xe in xe_list]
    for xe in xe_list:
        if xe.ndim != 2 or xe.shape[1] != 2:
            raise ValueError("All point sets must be 2D with shape (n_i, 2).")

    # Determine hulls if needed
    if hulls is None:
        hulls = []
        for xe in xe_list:
            if xe.shape[0] < 3:  # not enough points for a hull
                hulls.append(None)
            else:
                hulls.append(ConvexHull(xe))

    # Prepare figure
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Plot each convex hull
    for idx, (xe, hull) in enumerate(zip(xe_list, hulls)):
        color = f"C{idx % 10}"

        # Plot points
        ax.scatter(xe[:, 0], xe[:, 1], s=20, color=color)

        # Draw the hull polygon
        if hull is not None:
            verts = xe[hull.vertices]
            verts_closed = np.vstack([verts, verts[0]])

            ax.fill(
                verts_closed[:, 0],
                verts_closed[:, 1],
                facecolor=color,
                edgecolor=color,
                alpha=hull_alpha,
                linewidth=2,
            )

            # Bounding rectangle
            if show_bounds:
                min_bound = getattr(hull, "min_bound", xe.min(axis=0))
                max_bound = getattr(hull, "max_bound", xe.max(axis=0))
                xmin, ymin = min_bound
                xmax, ymax = max_bound

                rect_x = [xmin, xmax, xmax, xmin, xmin]
                rect_y = [ymin, ymin, ymax, ymax, ymin]
                ax.plot(rect_x, rect_y, "--", linewidth=1.5, color=color)

    # ax.set_aspect("equal", "box")
    ax.set_xlabel("x")
    ax.set_ylabel("y")

    if show:
        plt.show()

    return ax



from __future__ import annotations
from typing import *
from .intervals import Interval
import scipy.stats as sps
from .ecdf import transform_eCDF_bundle, get_ecdf, eCDF_bundle


def imprecise_ecdf_sps(s: Interval) -> tuple[eCDF_bundle, eCDF_bundle]:
    """empirical cdf for interval valued data

    caveat:
        with the use of `sps.ecdf`, the probability value does not start from 0.

    returns:
        - left and right cdfs
        - pbox
    """
    b_l = transform_eCDF_bundle(sps.ecdf(s.lo))
    b_r = transform_eCDF_bundle(sps.ecdf(s.hi))

    return b_l, b_r


def imprecise_ecdf(
    s: Interval, output_type: str = "ecdf"
) -> tuple[eCDF_bundle, eCDF_bundle]:
    """Empirical cdf for interval-valued data set

    args:
        s (Interval): interval valued data set
        output_type (str): the output type, either "pbox" or "ecdf"

    returns:
        - A tuple of left and right bounding cdf

    example:
        >>> rng = np.random.default_rng(seed=42)
        >>> precise_data = rng.normal(0, 1, 10)  # precise data case
        >>> impre_data = pba.I(lo = precise_data -0.5, hi = precise_data + 0.5)
        >>> l, r = imprecise_ecdf(s=impre_data, output_type="ecdf")
    """
    from .pbox_abc import pbox_from_ecdf_bundle

    b_l = eCDF_bundle(*get_ecdf(s.lo))
    b_r = eCDF_bundle(*get_ecdf(s.hi))

    if output_type == "pbox":
        return pbox_from_ecdf_bundle(b_l, b_r)
    elif output_type == "ecdf":
        return b_l, b_r
    else:
        raise ValueError("output_type must be either 'pbox' or 'ecdf'")



from __future__ import annotations
from typing import TYPE_CHECKING
from scipy.stats import qmc
import numpy as np

if TYPE_CHECKING:
    from ..pba.intervals import Interval

"""some helper funcs and classes for the epistemic space"""


class EpistemicDomain:
    """Representation of the epistemic space which are indeed bounds of each dimension

    This class provides a set of handy functions to work with epistemic uncertainty in the form of bounds.
    It will be useful for tasks such as propagation or optimization where epistemic uncertainty is involved.

    args:
        vars: a set of Interval variables

    tip:
        Recommended to use for optimisation tasks where the design bounds can be quickly specified with the ``toOptBounds()`` method.

    .. seealso::

        :class:`pyuncertainnumber.src.pyuncertainnumber.opt.bo` : Bayesian optimisation class.

        :class:`pyuncertainnumber.src.pyuncertainnumber.opt.ga` : Genetic algorithm class.

    example:
        >>> from pyuncertainnumber import pba
        >>> e = EpistemicDomain(pba.I(-1, 3), pba.I(5, 9))

        >>> # convert the epistemic space to bounds for the optimizer
        >>> e.toOptBounds(method='GA')  # `varbound` for genetic algorithm
        >>> e.toOptBounds(method='BO')  # `xc_bounds` for Bayesian optimisation

        >>> # perform lhs sampling on the epistemic space
        >>> sample = e.lhs_sampling(1000)
    """

    def __init__(self, *vars: Interval):
        from ..pba.intervals.intervalOperators import make_vec_interval, parse_bounds

        try:
            self.vec_interval = make_vec_interval(vars)
        except Exception as e:
            self.vec_interval = parse_bounds(*vars)

    def lhs_sampling(self, n_samples: int):
        """perform lhs sampling on the epistemic space"""
        Xc_sampler = qmc.LatinHypercube(d=len(self.vec_interval))
        l_bounds = self.vec_interval.lo
        u_bounds = self.vec_interval.hi

        base_sample = Xc_sampler.random(n=n_samples)
        return qmc.scale(base_sample, l_bounds, u_bounds)

    def lhs_plus_endpoints(self, n_samples: int):
        """perform lhs sampling on the epistemic space and add endpoints"""
        sample = self.lhs_sampling(n_samples)
        endpoints = self.to_GA_varBounds().T
        combined_sample = np.vstack((sample, endpoints))
        return combined_sample

    def bound_rep(self):
        """return the bounds (vec or matrix) of the epistemic space"""
        return self.vec_interval

    def toOptBounds(self, method: str):
        """convert the epistemic space to bounds for the optimizer

        args:
            method (str): the optimization method to use, e.g. 'BayesOpt', 'GA'

        returns:
            the bounds of the design varibale used for the optimisation method
        """
        if method == "BO":
            return self.to_BayesOptBounds()
        elif method == "GA":
            return self.to_GA_varBounds()
        else:
            raise ValueError(f"Unknown optimization method: {method}")

    def to_GA_varBounds(self) -> np.ndarray:
        """convert the epistemic space to bounds for the genetic algorithm optimizer"""
        varbound = self.vec_interval.to_numpy()
        if varbound.ndim == 1:
            varbound = varbound.reshape(-1, 2)
        return varbound

    def to_BayesOptBounds(self, func_signature="vectorisation") -> dict:
        """convert the epistemic space to bounds for the Bayesian optimisation optimizer"""

        # new_dict = self.__dict__.copy()
        # new_dict = {k: tuple(v) for k, v in new_dict.items()}
        # return new_dict
        if func_signature == "arguments":
            return {f"x{i}": tuple(r) for i, r in enumerate(self.to_GA_varBounds())}
        elif func_signature == "vectorisation" or func_signature == "iterable":
            return self.to_GA_varBounds()
        else:
            return self.to_GA_varBounds()



import numpy as numpy
import matplotlib as matplotlib
import matplotlib.pyplot as pyplot
from matplotlib import gridspec
from .peeling import peeling_to_structure
from .fuzzy import samples_to_fuzzy_projection
from numpy.typing import NDArray
import matplotlib.pyplot as plt

FONTSIZE = 22

FIGSIZE = {
    "small": (7, 7),
    "medium": (15, 15),
    "large": (20, 20),
}

color_dict = {
    "cyan": (0.0, 0.99, 0.99, 1),
    "yellow": (0.95, 0.95, 0, 1),
    "green": (0.0, 0.5, 0.0, 1),
    "orange": (1, 0.5, 0, 1),
    "red": (1, 0, 0, 1),
    "grey": (0.4, 0.4, 0.4, 0.8),
    "gray": (0.4, 0.4, 0.4, 0.8),
    "lightgrey": (0.7, 0.7, 0.7, 1),
    "blue": (0, 0, 0.8, 1),
    "blue2": (0, 0.4, 0.9, 1),
    "blue3": (0.5, 0.5, 0.8, 1),
    "black": (0, 0, 0, 1),
}


def c_(k, alpha=1):
    if k is None:
        t = list(color_dict["blue2"])
    elif type(k) is str:
        t = list(color_dict[k])
    else:
        if len(k) == 3:
            t = list(k) + [alpha]
        elif len(k) == 4:
            t = list(k)
    t[3] = alpha
    return tuple(t)


# cmaps = {('Perceptually Uniform Sequential',\
#             ['viridis', 'plasma', 'inferno', 'magma', 'cividis']),
#          ('Sequential',
#             ['Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
#             'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
#             'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
#          ('Sequential (2)', [
#             'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
#             'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
#             'hot', 'afmhot', 'gist_heat', 'copper']),
#          ('Diverging', [
#             'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
#             'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']),
#          ('Cyclic', ['twilight', 'twilight_shifted', 'hsv']),
#          ('Qualitative', [
#             'Pastel1', 'Pastel2', 'Paired', 'Accent',
#             'Dark2', 'Set1', 'Set2', 'Set3',
#             'tab10', 'tab20', 'tab20b', 'tab20c']),
#          ('Miscellaneous', [
#             'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
#             'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
#             'gist_rainbow', 'rainbow', 'jet', 'turbo', 'nipy_spectral',
#             'gist_ncar'])}

cmap = matplotlib.cm.get_cmap("tab20c")
COLORMAP = cmap.colors


def breakout(n):  # use this to determine the grid size of subplots
    def factor(n):
        t = numpy.arange(2, n, 1)
        return t[
            n % t == 0
        ]  # returns list of cofactors of an integer `n`, empty if integer `n` is prime.

    if n == 1:
        return 1, 1
    elif n == 2:
        return 1, 2
    elif n == 3:
        return 1, 3
    f = factor(n)
    while len(f) < 1:
        n += 1
        f = factor(n)
    l = len(f)
    if l == 1:
        a, b = f[0], f[0]
    elif l == 2:
        a, b = f[0], f[1]
    if l > 2:
        b = f[l // 2]
        a = n // b
    return a, b


def plot_peeling(
    x: NDArray, a, b, p=None, axes3d=False, figsize="medium", grid=True, label="X"
):
    """Plotting function for data peeling results.

    Args:
        x (NDArray): data set of iid observations
        a: sequence of subindices for each level
        b: sequence of boxes or enclosing sets
        p: upper violation probability (membership value)
    """
    x = numpy.asarray(x, dtype=float)
    n, d = x.shape
    if d == 1:
        pass
    elif d == 2:
        plot_peeling_nx2(x, a, b, p=p, figsize=figsize, grid=grid, label=label)
        return None
    elif d == 3:
        if axes3d:
            plot_peeling_3d(x, a, b, p=p, figsize=figsize, grid=grid, label=label)
        return None
    plot_peeling_nxd(
        x, a, b, p=p, figsize=figsize, grid=grid, label=label
    )  # scattermatrix with fuzzy marginals
    return None


def plot_one_fuzzy_grad(
    a_fuzzy,
    p=None,
    data=None,
    ax=None,
    figsize=None,
    grid=None,
    color=None,
    baseline_alpha=0.4,
    linewidth=0.1,
    colormap=None,
    xlabel=r"$X$",
    ylabel=r"$1-\delta$",
    flip=False,
):
    a = numpy.asarray(a_fuzzy, dtype=float)
    l = a.shape[0]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    fig = None
    if ax is None:
        fig, ax = pyplot.subplots(figsize=figsize)
    if p is None:
        p = numpy.arange(1 / l, 1, 1 / l)  # p = 1/n
    p = [0] + list(p)  # add ficticius zero to enable the blox plot
    alpha = numpy.linspace(baseline_alpha, 1, num=l)
    for i in range(l):
        x = [a[i, 0], a[i, 1]]
        y1 = [p[i], p[i]]  # y1=[p*i,p*i]
        y2 = [p[i + 1], p[i + 1]]  # y2=[p*(i+1),p*(i+1)]
        if colormap is not None:
            lenc = len(colormap)
            if l > lenc:
                color = colormap[i % lenc]
            else:
                color = colormap[i * (lenc // l)]
        if flip:
            ax.fill_betweenx(
                y=x,
                x1=y1,
                x2=y2,
                facecolor=c_(color, alpha=alpha[i]),
                edgecolor=c_("grey", alpha=0.8),
                linewidth=linewidth,
            )
        else:
            ax.fill_between(
                x=x,
                y1=y1,
                y2=y2,
                facecolor=c_(color, alpha=alpha[i]),
                edgecolor=c_("grey", alpha=0.8),
                linewidth=linewidth,
            )
        if data is not None:
            inside = (data >= x[0]) & (data <= x[1])
            data_level = numpy.sort(data[inside])
            if flip:
                ax.scatter(
                    2 * [p[i + 1]],
                    [data_level[0], data_level[-1]],
                    color="grey",
                    alpha=0.5,
                    s=10,
                )
            else:
                ax.scatter(
                    [data_level[0], data_level[-1]],
                    2 * [p[i + 1]],
                    color="grey",
                    alpha=0.5,
                    s=10,
                )
    if flip:
        ax.plot(
            2 * [p[i + 1]], [x[0], x[1]], color=c_("black", alpha=0.9), linewidth=0.5
        )
    else:
        ax.plot(
            [x[0], x[1]], 2 * [p[i + 1]], color=c_("black", alpha=0.9), linewidth=0.5
        )
    if data is not None:
        if flip:
            ax.scatter(len(data) * [0], data, color="grey", alpha=0.5, s=30, marker="s")
        else:
            ax.scatter(data, len(data) * [0], color="grey", alpha=0.5, s=30, marker="s")
    if grid is not None:
        ax.grid()
    ax.set_xlabel(xlabel, fontsize=FONTSIZE)
    ax.set_ylabel(ylabel, fontsize=FONTSIZE)
    return fig, ax


def plot_box(
    box2d,
    ax=None,
    figsize=(10, 10),
    facecolor=None,
    edgecolor=None,
    alpha=None,
    label=None,
    zorder=None,
    grid=True,
):
    a, b = box2d[0], box2d[1]
    fig = None
    if ax is None:
        fig, ax = pyplot.subplots(figsize=figsize)
    ax.fill_between(
        x=[a[0], a[1]],
        y1=[b[0], b[0]],
        y2=[b[1], b[1]],
        facecolor=facecolor,
        edgecolor=edgecolor,
        alpha=alpha,
        label=label,
        zorder=zorder,
    )
    if grid:
        ax.grid()
    return fig, ax


def plot_peeling_nx2(
    X,
    a,
    b,
    p: list = None,
    max_level: int = None,
    label="X",
    grid=True,
    savefig: str = None,
    figsize=None,
    baseline_alpha=0.075,
):
    b = numpy.asarray(b, dtype=float)
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    fig, axe = pyplot.subplots(
        7, 7, figsize=figsize, constrained_layout=False, sharex=True
    )
    gs = axe[0, 6].get_gridspec()
    fig.subplots_adjust(wspace=10)
    for j in range(7):
        for axi in axe[:, j]:
            axi.remove()
    axbig = fig.add_subplot(gs[2:, :5])
    ax_up = fig.add_subplot(gs[0:2, :5])
    ax_ri = fig.add_subplot(gs[2:, 5:])

    ax_up.set_ylim(0, 1)
    ax_ri.set_xlim(0, 1)

    axbig.tick_params(
        direction="out", length=6, width=0.5, labelsize=16
    )  # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.tick_params.html
    ax_up.tick_params(
        direction="out", length=6, width=0.5, labelsize=16
    )  # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.tick_params.html
    ax_ri.tick_params(
        direction="out", length=6, width=0.5, labelsize=16
    )  # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.tick_params.html

    axbig.scatter(X[:, 0], X[:, 1], zorder=10, marker="s")

    if max_level is None:
        L = len(a)

    alpha = baseline_alpha

    for i in range(L):
        _ = plot_box(
            b[i],
            ax=axbig,
            facecolor=c_("blue2", alpha=alpha),
            edgecolor=c_("red", alpha=alpha),
            zorder=0,
            grid=False,
        )
        axbig.scatter(X[a[i], 0], X[a[i], 1], zorder=11, marker="s", color="grey")

    _ = plot_box(
        b[L - 1],
        ax=axbig,
        facecolor=c_("blue2", alpha=alpha),
        edgecolor=c_("red", alpha=alpha),
        zorder=0,
        grid=False,
    )
    axbig.scatter(X[a[L - 1], 0], X[a[L - 1], 1], zorder=12, marker="s", color="red")

    # alpha = 0.4
    _ = plot_one_fuzzy_grad(
        b[:, 0, :], data=X[:, 0], p=p, ax=ax_up, color="blue2", baseline_alpha=alpha
    )
    _ = plot_one_fuzzy_grad(
        b[:, 1, :],
        data=X[:, 1],
        p=p,
        ax=ax_ri,
        color="blue2",
        baseline_alpha=alpha,
        flip=True,
    )

    if grid:
        axbig.grid()
        ax_up.grid()
        ax_ri.grid()

    axbig.set_aspect("auto")
    axbig.set_xlabel(r"$" + label + "_1$", fontsize=FONTSIZE)
    axbig.set_ylabel(r"$" + label + "_2$", fontsize=FONTSIZE)
    ax_up.set_xlabel(r"$" + label + "_1$", fontsize=FONTSIZE)
    ax_ri.set_ylabel(r"$" + label + "_2$", fontsize=FONTSIZE)
    ax_up.set_ylabel(r"$1-\delta$", fontsize=FONTSIZE)
    ax_ri.set_xlabel(r"$1-\delta$", fontsize=FONTSIZE)

    fig.tight_layout()
    if savefig is not None:
        fig.savefig(savefig)
    pass


def plot_peeling_3d(x, a, b):
    pass


def plot_peeling_nxd(
    x,
    a,
    b,
    fx=None,
    p: list = None,
    figsize=None,
    aspect="auto",
    label="X",
    marker="s",
    markercolor="grey",
    boxcolor="blue2",
    grid=True,
    baseline_alpha=0.075,
):
    n, d = x.shape
    l = len(b)
    if (p is None) | (fx is None):
        fx, p = peeling_to_structure(a, b, kind="scenario", beta=0.01)
    labels = [r"$" + label + "_" + str(i + 1) + "$" for i in range(d)]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    boxcolor = c_(boxcolor, alpha=baseline_alpha)
    markercolor = c_(markercolor, alpha=0.5)
    fig = pyplot.figure(figsize=figsize)
    gs = gridspec.GridSpec(d, d, figure=fig)
    for i in range(d):
        for j in range(d):
            ax = pyplot.subplot(gs[j, i])
            if i == j:  # on diagonal
                _ = plot_one_fuzzy_grad(
                    fx[:, i, :],
                    data=x[:, i],
                    p=p,
                    ax=ax,
                    color=boxcolor,
                    baseline_alpha=baseline_alpha,
                )
                ax.set_xlabel(labels[i], fontsize=FONTSIZE)
                ax.set_ylabel(r"$1-\delta$", fontsize=FONTSIZE)
            else:
                ax.scatter(
                    x[:, i],
                    x[:, j],
                    color=markercolor,
                    marker=marker,
                    edgecolors="face",
                )
                for box in b:
                    _ = plot_box([box[i, :], box[j, :]], ax=ax, facecolor=boxcolor)
                ax.set_aspect(aspect)
                ax.set_xlabel(labels[i], fontsize=FONTSIZE)
                ax.set_ylabel(labels[j], fontsize=FONTSIZE)
            if grid:
                ax.grid()
            ax.tick_params(
                direction="out",
                length=6,
                width=2,
                colors="#5a5a64",
                grid_color="gray",
                grid_alpha=0.5,
                labelsize="x-large",
            )
    pyplot.tight_layout()
    return ax


def plot_scattermatrix(
    x,
    bins=10,
    GS=None,
    figsize=None,
    aspect="auto",
    color=None,
    marker="s",
    alpha=None,
    edgecolors="face",
    grid=True,
    label="X",
):
    n, d = x.shape
    labels = [r"$" + label + "_" + str(i + 1) + "$" for i in range(d)]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    if GS is None:
        fig = pyplot.figure(figsize=figsize)
        GS = gridspec.GridSpec(d, d, figure=fig)
    SUBAX = []
    for i in range(d):
        for j in range(d):
            ax = pyplot.subplot(GS[i, j])
            if i == j:
                ax.hist(x[:, i], bins=bins, density=True, color=color)
                # ax.yaxis.tick_right()
                ax.set_xlabel(labels[i], fontsize=FONTSIZE)
                ax.set_ylabel(f"{i+1}/N", fontsize=FONTSIZE)
            else:
                ax.scatter(
                    x[:, j],
                    x[:, i],
                    color=color,
                    marker=marker,
                    alpha=alpha,
                    edgecolors=edgecolors,
                )
                ax.set_aspect(aspect)
                ax.set_xlabel(labels[j], fontsize=FONTSIZE)
                ax.set_ylabel(labels[i], fontsize=FONTSIZE)
            SUBAX.append(ax)
    for ax in SUBAX:
        if grid:
            ax.grid(True)
        ax.tick_params(
            direction="out",
            length=6,
            width=2,
            colors="#5a5a64",
            grid_color="gray",
            grid_alpha=0.5,
            labelsize="x-large",
        )
    pyplot.tight_layout()


def plot_fuzzy(
    fuzzy,
    p=None,
    data=None,
    ax=None,
    figsize=None,
    grid=False,
    color=None,
    baseline_alpha=0.4,
    linewidth=0.1,
    colormap=None,
    xlabel=None,
    ylabel=r"$1-\delta$",
    flip=False,
):
    """
    fuzzy: An (l,d_,2) array with projections
    """
    l, d_, _ = fuzzy.shape
    s1, s2 = breakout(d_)
    if xlabel is None:
        labels = [r"$X_" + str(i + 1) + "$" for i in range(d_)]
    if type(xlabel) == list:
        labels = xlabel
    if type(xlabel) == str:
        labels = [r"$" + str(xlabel) + "_" + str(i + 1) + "$" for i in range(d_)]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    fig, axes = pyplot.subplots(nrows=s1, ncols=s2, figsize=figsize)
    axes = numpy.matlib.repmat(axes, 1, 1)
    k = 0
    for i in range(s1):
        for j in range(s2):
            if k < d_:
                ax = axes[i, j]
                plot_one_fuzzy_grad(
                    fuzzy[:, k, :],
                    p=p,
                    ax=ax,
                    color=color,
                    baseline_alpha=baseline_alpha,
                    xlabel=labels[k],
                    ylabel=ylabel,
                )
                k += 1
                if grid:
                    ax.grid()
                ax.tick_params(
                    direction="out",
                    length=6,
                    width=2,
                    colors="#5a5a64",
                    grid_color="gray",
                    grid_alpha=0.5,
                    labelsize="x-large",
                )
    pyplot.tight_layout()
    pass


def plot_peeling_nxd_back(
    ux,
    c,
    p: list = None,
    figsize=None,
    aspect="auto",
    xlabel="X",
    ylabel=r"$1-\delta$",
    marker="s",
    markercolor="grey",
    boxcolor="blue2",
    colormap=None,
    grid=True,
    baseline_alpha=0.85,
):
    m, d_ = ux.shape
    l = len(c)
    fx = samples_to_fuzzy_projection(ux, c)
    if colormap is None:
        colormap = COLORMAP
    if p is None:
        p = numpy.arange(1 / l, 1 + 1 / l, 1 / l)  # p = 1/l
    if xlabel is None:
        labels = [r"$X_" + str(i + 1) + "$" for i in range(d_)]
    if type(xlabel) == list:
        labels = xlabel
    if type(xlabel) == str:
        labels = [r"$" + str(xlabel) + "_" + str(i + 1) + "$" for i in range(d_)]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif type(figsize) == str:
        figsize = FIGSIZE[figsize]
    boxcolor = c_(boxcolor, alpha=baseline_alpha)
    markercolor = c_(markercolor, alpha=0.5)
    fig = pyplot.figure(figsize=figsize)
    gs = gridspec.GridSpec(d_, d_, figure=fig)
    for i in range(d_):
        for j in range(d_):
            ax = pyplot.subplot(gs[j, i])
            if i == j:  # on diagonal
                _ = plot_one_fuzzy_grad(
                    fx[:, i, :],
                    p=p,
                    ax=ax,
                    colormap=colormap,
                    baseline_alpha=baseline_alpha,
                )
                ax.set_xlabel(labels[i], fontsize=FONTSIZE)
                ax.set_ylabel(ylabel, fontsize=FONTSIZE)
            else:
                ax.scatter(ux[:, i], ux[:, j], color="gray", alpha=0.1)
                for k in range(len(c)):
                    if len(cmap.colors) < len(c):
                        colors = cmap.colors[k % len(cmap.colors)]
                    else:
                        colors = cmap.colors[k * (len(cmap.colors) // len(c))]
                    ax.scatter(
                        ux[c[k], i],
                        ux[c[k], j],
                        alpha=0.9,
                        s=20,
                        marker="s",
                        color=colors,
                    )
                ax.set_aspect(aspect)
                ax.set_xlabel(labels[i], fontsize=FONTSIZE)
                ax.set_ylabel(labels[j], fontsize=FONTSIZE)
            if grid:
                ax.grid()
            ax.tick_params(
                direction="out",
                length=6,
                width=2,
                colors="#5a5a64",
                grid_color="gray",
                grid_alpha=0.5,
                labelsize="x-large",
            )
    pyplot.tight_layout()
    pass


# * --------- refactor

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec


def _draw_peeling_cell(
    ax,
    x,
    b,
    fx,
    p,
    i,
    j,
    labels,
    aspect="auto",
    marker="s",
    markercolor="grey",
    boxcolor="blue2",
    grid=True,
    baseline_alpha=0.075,
):
    # match your styling
    boxcolor_rgba = c_(boxcolor, alpha=baseline_alpha)
    markercolor_rgba = c_(markercolor, alpha=0.5)

    if i == j:  # diagonal
        _ = plot_one_fuzzy_grad(
            fx[:, i, :],
            data=x[:, i],
            p=p,
            ax=ax,
            color=boxcolor_rgba,
            baseline_alpha=baseline_alpha,
        )
        ax.set_xlabel(labels[i], fontsize=FONTSIZE)
        ax.set_ylabel(r"$1-\delta$", fontsize=FONTSIZE)
    else:  # off-diagonal
        ax.scatter(
            x[:, i],
            x[:, j],
            color=markercolor_rgba,
            marker=marker,
            edgecolors="face",
        )
        for box in b:
            _ = plot_box([box[i, :], box[j, :]], ax=ax, facecolor=boxcolor_rgba)

        ax.set_aspect(aspect)
        ax.set_xlabel(labels[i], fontsize=FONTSIZE)
        ax.set_ylabel(labels[j], fontsize=FONTSIZE)

    if grid:
        ax.grid()

    ax.tick_params(
        direction="out",
        length=6,
        width=2,
        colors="#5a5a64",
        grid_color="gray",
        grid_alpha=0.5,
        labelsize="x-large",
    )

    return ax


def plot_peeling_nxd_all(
    x,
    a,
    b,
    fx=None,
    p: list = None,
    figsize=None,
    aspect="auto",
    label="X",
    marker="s",
    markercolor="grey",
    boxcolor="blue2",
    grid=True,
    baseline_alpha=0.075,
    return_axes=False,  # <-- new
):
    n, d = x.shape
    if (p is None) or (fx is None):
        fx, p = peeling_to_structure(a, b, kind="scenario", beta=0.01)

    labels = [r"$" + label + "_" + str(i + 1) + "$" for i in range(d)]
    if figsize is None:
        figsize = FIGSIZE["medium"]
    elif isinstance(figsize, str):
        figsize = FIGSIZE[figsize]

    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(d, d, figure=fig)

    axes = np.empty((d, d), dtype=object)

    for i in range(d):
        for j in range(d):
            ax = plt.subplot(gs[j, i])
            axes[j, i] = ax  # row=j, col=i (matches your gs[j,i])
            _draw_peeling_cell(
                ax=ax,
                x=x,
                b=b,
                fx=fx,
                p=p,
                i=i,
                j=j,
                labels=labels,
                aspect=aspect,
                marker=marker,
                markercolor=markercolor,
                boxcolor=boxcolor,
                grid=grid,
                baseline_alpha=baseline_alpha,
            )

    plt.tight_layout()
    return (fig, axes) if return_axes else ax  # keep backward compatible


def plot_peeling_one(
    x,
    a,
    b,
    i,
    j,
    fx=None,
    p: list = None,
    figsize=(4, 4),
    aspect="auto",
    label="X",
    marker="s",
    markercolor="grey",
    boxcolor="blue2",
    grid=True,
    baseline_alpha=0.075,
):
    n, d = x.shape
    if (p is None) or (fx is None):
        fx, p = peeling_to_structure(a, b, kind="scenario", beta=0.01)

    labels = [r"$" + label + "_" + str(k + 1) + "$" for k in range(d)]

    fig, ax = plt.subplots(figsize=figsize)
    _draw_peeling_cell(
        ax=ax,
        x=x,
        b=b,
        fx=fx,
        p=p,
        i=i,
        j=j,
        labels=labels,
        aspect=aspect,
        marker=marker,
        markercolor=markercolor,
        boxcolor=boxcolor,
        grid=grid,
        baseline_alpha=baseline_alpha,
    )
    fig.tight_layout()
    return fig, ax



from .intervals.number import Interval as I
from .intervals.methods import intervalise, load_interval_from_json

from .pbox_parametric import *
from .pbox_free import *
from .dss import DempsterShafer
from .dss import DempsterShafer as DSS
from .distributions import Distribution as D
from .distributions import Distribution
from .distributions import JointDistribution, ECDF
from .dependency import Dependency
from .context import dependency
from .pbox_abc import inspect_pbox, pbox_from_ecdf_bundle
from pyuncertainnumber.pba.ecdf import get_ecdf
from .operation import convert
from .aggregation import *
from .pbox_free import KS_bounds
from .core import area_metric



from pyuncertainnumber.pba.utils import variance_bounds_via_lp
import pytest
from pyuncertainnumber import pba
import numpy as np

pbox = pba.normal([4, 20], 1)

x_grid = np.linspace(pbox.lo, pbox.hi, 50)


def test_function_runs_without_exception():

    try:
        variance_bounds_via_lp(
            q_a=pbox.left,
            p_a=pbox.p_values,
            q_b=pbox.right,
            p_b=pbox.p_values,
            x_grid=x_grid,
            n=200,
            mu_grid=101,
        )
    except Exception as e:
        pytest.fail(f"Function raised an exception: {e}")



import pytest
import jax.numpy as jnp
from pyuncertainnumber import taylor_expansion_method


def test_taylor_expansion_method():

    MEAN = jnp.array([3.0, 2.5])
    COV = jnp.array([[4, 0.3], [0.3, 0.25]])

    def bar(x):
        return x[0] ** 2 + x[1] + 3

    mu_, var_ = taylor_expansion_method(func=bar, mean=MEAN, cov=COV)
    assert mu_ == 18.5
    assert var_ == pytest.approx(179.849, rel=1e-1)



from pyuncertainnumber import pba
from pyuncertainnumber import UN


def test_single_parameter_construction():
    # single-parameter distribution
    a = pba.pareto(2.62)
    b_d = pba.D("pareto", 2.62)
    b = b_d.to_pbox()
    c_UN = UN(essence="distribution", distribution_parameters=["pareto", 2.62])
    c = c_UN.to_pbox()

    assert a == b and b == c and a == c, "Single-parameter construction problem"



import numpy as np

from pyuncertainnumber import pba
from pyuncertainnumber.pba.pbox_abc import Staircase
from ..pba.dependency import Dependency
from ..pba.distributions import JointDistribution

"""independence of input pboxes are assumed to get started"""


def middle_pinch(pbox_list: list):
    """Perform middle pinch on a list of p-boxes.

    args:
        pbox_list (list): A list of marginal p-boxes to be processed.
    """
    from pyuncertainnumber.pba.pbox_abc import Staircase

    def middle_pinch_function(pbox):
        q = (pbox.left + pbox.right) / 2
        return Staircase(left=q, right=q)

    return [middle_pinch_function(pbox) for pbox in pbox_list]


class SurrogatePropagation:
    """Surrogate propagation for uncertain numbers: (Pbox for now)

    args:

        vars (uncertain number or constructs):
            input pboxes

        surrogate_model:
            supported surrogate models in `probabilistic_model_archive`, {IPM, GP, ...}

        N1 (int): number of experimental design for level-1 meta-modelling

        N2 (int): number of experimental design for level-2 meta-modelling

        f (callable): True DGM computational model;

    note:
        - notable examples: GP or IPM to propagate p-boxes for imprecise reliability analysis;
        - Independence of input p-boxes is assumed to get started.
        - the two-level framework is used.
    """

    def __init__(
        self,
        vars,
        f,
        surrogate_model,
        N1=100,
        N2=200,
        dependency: Dependency = None,
        auxiliary_input_distribution=None,
        random_state=None,
    ):
        self.vars = vars
        self.f = f
        self.surrogate_model = surrogate_model
        self.N1 = N1
        self.N2 = N2
        self.dependency = dependency
        self.auxiliary_input_distribution = (
            middle_pinch(self.vars)
            if auxiliary_input_distribution is None
            else auxiliary_input_distribution
        )
        self.random_state = random_state
        ### add the level-1 and level-2 logic herein
        # self.level_1_meta_modelling()
        # self.level_2_meta_modelling()
        # self.use_l2_get_response()

    def level_1_meta_modelling(self):
        # Implement level-1 meta-modelling here
        from pyuncertainnumber.pba.intervals.number import interval_degenerate

        alpha_l1 = self.dependency.u_sample(
            self.N1, random_state=self.random_state
        )  # (N1, d)

        X_tilda_input = []
        for i, v in enumerate(self.auxiliary_input_distribution):
            arr = interval_degenerate(v.alpha_cut(alpha_l1[:, i]))
            X_tilda_input.append(arr)
        X_tilda_input = np.array(X_tilda_input).T  # (N1, d)

        y_tilda_output = self.f(X_tilda_input)  # get real y_obs

        return X_tilda_input, y_tilda_output

        ### to train a level-1 surrogate model ###
        # self.l1_model = self.surrogate_model.fit(X_tilda_input, y_tilda_output)

    def level_2_meta_modelling(self, trained_surrogate):
        """Level-2 meta-modelling.

        This step involves interval propagation accomplished by optimisation, therefore it takes some time to execute.

        note:
            mapping between ɑ and y_low, ɑ and y_high. an IPM model can achieve the mapping using just one model.
        """
        from pyuncertainnumber import b2b

        ndim = len(self.vars)
        prob_proxy_input = self.dependency.u_sample(
            self.N2, random_state=self.random_state
        )  # (N2, d)
        y_lo = np.zeros(len(prob_proxy_input))
        y_hi = np.zeros(len(prob_proxy_input))

        for i, row in enumerate(prob_proxy_input):
            x_domain = [
                v.alpha_cut(a) for v, a in zip(self.vars, row)
            ]  # yield a list of intervals
            # TODO: fix the issue of being too slow when using surrogate model for optimisation
            response_y_itvl = b2b(
                vars=x_domain,
                func=self.f,  # temporarily use true DGM to replace surrogate
                # func=trained_surrogate,
                interval_strategy="ga",
            )

            y_lo[i] = response_y_itvl.lo
            y_hi[i] = response_y_itvl.hi

        return prob_proxy_input, y_lo, y_hi
        # with (ɑ, y_lo) and (ɑ, y_hi), train 2 level-2 surrogate models

    def l2_get_response(
        self, meta_model_lo=None, meta_model_hi=None, ipm_model=None, n=1000_000
    ):
        """with l2 model, predict the vector of probability to get the quantile"""

        if not (
            (meta_model_lo is not None and meta_model_hi is not None)
            or (ipm_model is not None)
        ):
            raise ValueError(
                "You must provide either both meta_model_lo and meta_model_hi, or ipm_model."
            )

        # Xnew
        alpha_new_ = self.dependency.u_sample(n=n, random_state=self.random_state)

        if ipm_model is not None:
            mean_hi, mean_lo = ipm_model(
                alpha_new_
            )  # not technically mean_lo but reuse the name for convenience
        else:
            mean_lo = meta_model_lo(X_new=alpha_new_)
            mean_hi = meta_model_hi(X_new=alpha_new_)

        q_l, _ = pba.get_ecdf(np.sort(mean_lo))
        q_r, _ = pba.get_ecdf(np.sort(mean_hi))
        response = Staircase(left=q_l, right=q_r)
        return response



from .peeling import (data_peeling_algorithm,data_peeling_backward,peeling_to_structure,uniform)
from .plots import (plot_peeling,plot_peeling_nxd,plot_peeling_nxd_back,plot_peeling_nx2,plot_scattermatrix)
from .fuzzy import (samples_to_fuzzy_projection,boxes_to_fuzzy_projection,coverage_samples)
from .examples import (pickle_load,pickle_dump,banana_data,banana_model)


import pyuncertainnumber as pun
import numpy as np
from pyuncertainnumber import pba
from pyuncertainnumber import UncertainNumber as UN
from pyuncertainnumber.pba.pbox_abc import Pbox
from pyuncertainnumber.pba.aggregation import stochastic_mixture


# *  ---------------------construction---------------------*#
def test_pun_pb_constuction():
    assert isinstance(
        pun.norm([2, 3], [0.1]), UN
    ), "Failed to construct a Pbox-type UncertainNumber object at UN level"

    assert isinstance(
        pun.I([2, 3]), UN
    ), "Failed to construct an Interval-type UncertainNumber object at UN level"


# *  ---------------------aggregation---------------------*#


def test_interval_aggregation():
    """Interval aggregation as in expert opinions"""
    lower_endpoints = np.random.uniform(-0.5, 0.5, 7)
    upper_endpoints = np.random.uniform(0.5, 1.5, 7)
    m_weights = [0.1, 0.1, 0.25, 0.15, 0.1, 0.1, 0.2]
    # a list of nInterval objects
    nI = [pba.I(*couple) for couple in zip(lower_endpoints, upper_endpoints)]

    pbox_mix = stochastic_mixture(*nI, weights=m_weights)
    print("the result of the mixture operation")
    assert isinstance(
        pbox_mix, Pbox
    ), "Failed to aggregate weights intervals expert opinions into Pbox objects"


# *  ---------------------arithmetic---------------------*#


def test_pba_arithmetic():
    """pba level"""
    a = pba.I([2, 3])  # an interval
    # _ = a.display(style="band", title="Interval [2,3]")
    b = pba.norm(0, 1)  # a precise distribution
    # _ = b.display(title="$N(0, 1)$")
    t = a + b
    assert isinstance(t, Pbox), "Failed to add a Pbox object at PBA level"


def test_UN_arithmetic():
    """UN level"""
    a = pun.norm(0, 1)
    b = pun.norm(2, 3)
    t = a + b

    assert isinstance(
        t, UN
    ), "Failed to add a Pbox-type UncertainNumber object at UN level"



from __future__ import annotations
from typing import TYPE_CHECKING
import functools
from .uncertainNumber import UncertainNumber
from .stats import *

if TYPE_CHECKING:
    from pyuncertainnumber import Interval

""" to store core functions and decorators (underway)"""


def makeUN(func):
    """return from construct a Uncertain Number object"""

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        construct = func(*args, **kwargs)
        return UncertainNumber.fromConstruct(construct)

    return wrapper_decorator



"""
:#######################################################
: Intervals Library originally created by Marco de Angelis
: Continuesly developed and refactored by Leslie Yu Chen
:#######################################################
"""

# Sequence: # Must have __len__() and __getitem__(). Ex.: Tuple, List, Range
# Sized: # It suffices to have len()
# Iterable: # Must have __iter__() and __next__(). Ex.: Dict, Set, Tuple, List, numpy.array
# from intervals.methods import (lo,hi,width,rad,mag,straddlezero,isinterval)

from __future__ import annotations
from fileinput import filename
from typing import Optional, Any, Union
from pathlib import Path
import json5
import numpy as np
import numpy
from numpy import ndarray, asarray, stack, transpose, zeros
from scipy.stats import qmc
import matplotlib.pyplot as plt
from .utils import safe_asarray
from ..mixins import NominalValueMixin
from .arithmetic import multiply, divide


MACHINE_EPS = 7.0 / 3 - 4.0 / 3 - 1

NUMERIC_TYPES = {
    "int",
    "float",
    "complex",  # Python numbers
    "int8",
    "int16",
    "int32",
    "int64",
    "intp",  # Numpy integers
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "uintp",  # Numpy unsigned integers
    "float16",
    "float32",
    "float64",
    "float_",  # Numpy floats and doubles
    "complex64",
    "complex128",
    "complex_",
}  # Numpy complex floats and doubles

INTEGERS = {
    "int",
    "int8",
    "int16",
    "int32",
    "int64",
    "intp",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "uintp",
}
# FLOATS =            {'float','float16','float32','float64','float_'}


def show(x: Interval) -> str:
    if len(x) == 0:
        return f"[{x.lo},{x.hi}]"
    elif len(x.shape) == 1:
        return "\n".join([f"[{xi.lo},{xi.hi}]" for xi in x])  # vector of intervals
    elif len(x.shape) == 2:
        n, d = x.shape
        return "\n".join(
            [" ".join([f"{xi.val}" for xi in x[i, :]]) for i in range(n)]
        )  # matrix of intervals
    else:
        return f"{x.val}"


class Interval(NominalValueMixin):
    """Interval is the main class"""

    def __init__(
        self,
        lo: Union[float, ndarray],
        hi: Optional[Union[float, ndarray]] = None,
        do_heavy_checks: bool = True,
    ) -> None:
        self._lo = safe_asarray(lo)
        if hi is None:
            hi = self._lo.copy()
        self._hi = safe_asarray(hi)
        if do_heavy_checks:
            self.run_heavy_checks()

    def run_heavy_checks(self):
        """Run heavy checks on the interval object"""
        assert np.all(
            self._lo <= self._hi
        ), "low larger than high, needed to invert the interval"
        self.__shape = self._lo.shape

    def __repr__(self):  # return
        return show(self)

    def __str__(self):  # print
        return show(self)

    def __len__(self):
        if self.unsized:
            return 0  # interval object is not sized, perhaps return an error: TypeError: len() of unsized object
        else:
            return self._lo.shape[0]

    def __iter__(self):  # https://realpython.com/introduction-to-python-generators/
        lo_iter, hi_iter = numpy.nditer(self.lo), numpy.nditer(self.hi)
        while True:
            try:
                yield Interval(lo=next(lo_iter), hi=next(hi_iter))
            except StopIteration:
                break
        pass

    def __contains__(self, item):
        """Check if an item is enclosed within the interval.

        example:
            >>> i = Interval(1,3)
            >>> 2 in i
            True
            >>> 4 in i
            False
        """
        return np.all((item >= self.lo) & (item <= self.hi))

    def __next__(self):
        pass

    def __getitem__(self, i: Union[int, slice]):  # make class indexable
        return Interval(lo=self._lo[i], hi=self._hi[i])

    # * -------------- METHODS -------------- *#
    def to_numpy(self) -> np.ndarray:
        """transform interval objects to numpy arrays"""
        if self.scalar:
            return np.array([self.lo.item(), self.hi.item()])
        else:
            return np.asarray((self.lo, self.hi)).T

    def to_pbox(self):
        from ..pbox_abc import Staircase
        from ..params import Params

        return Staircase(
            left=np.repeat(self.lo, Params.steps),
            right=np.repeat(self.hi, Params.steps),
            mean=self,
            var=Interval(0, (self.hi - self.lo) * (self.hi - self.lo) / 4),
        )

    def lhs_sample(self, n) -> np.ndarray:
        """LHS sampling within the interval

        args:
            n: number of samples
        """
        if self.is_scalar:
            sampler = qmc.LatinHypercube(d=1)
        else:
            sampler = qmc.LatinHypercube(d=self.__len__())
        sample = sampler.random(n=n)
        sample = qmc.scale(sample, self.lo, self.hi)
        return sample

    def endpoints_lhs_sample(self, n) -> np.ndarray:
        """LHS sampling within the interval plus the endpoints

        args:
            n: number of samples
        """
        lhs_sample = self.lhs_sample(n)
        endpoints = np.array([self.lo, self.hi])
        if self.is_scalar:  # lhs_sample ~ (n, 1)
            return np.concatenate((lhs_sample, endpoints[:, np.newaxis]))
        else:
            return np.vstack((endpoints, lhs_sample))

    def plot(self, ax=None, **kwargs):
        p = self.to_pbox()
        if ax is None:
            fig, ax = plt.subplots()
        p.plot(ax=ax, **kwargs)

    def display(self):
        self.plot()
        plt.show()

    def is_degenerate(self) -> bool:
        """Check if the interval is degenerate (i.e., has zero width)."""
        if np.all(self._lo == self._hi):
            return True
        return False

    def _compute_nominal_value(self):
        return self.mid

    def ravel(self):
        """Return a flattened (1D) interval object for multi-dimensional intervals

        example:
            >>> A = np.random.rand(200, 200, 2)
            >>> i = pba.intervalise(A)
            >>> print(i.shape)
            >>> i2 = i.ravel()
            >>> print(i2.shape)
        """
        oned_cc = Interval(self.lo.ravel(), self.hi.ravel())
        return oned_cc

    @property
    def lo(self) -> Union[ndarray, float]:
        return self._lo

    # if len(self.shape)==0: return self._lo
    # return self._lo # return transpose(transpose(self.__val)[0]) # from shape (3,7,2) to (2,7,3) to (3,7)

    @property
    def hi(self) -> Union[ndarray, float]:
        return self._hi

    @property
    def left(self):
        return self.lo

    @property
    def right(self):
        return self.hi

    @property
    def width(self):
        return width(self)

    @property
    def rad(self):
        """half width"""
        return rad(self)

    @property
    def mid(self):
        return mid(self)

    @property
    def unsized(self):
        if (len(self._hi.shape) > 0) | (len(self._hi.shape) > 0):
            return False
        else:
            return True

    @property
    def val(self):
        """seemingly equivalent to `self.to_numpy()`"""
        if self.unsized:
            return asarray([self._lo, self._hi], dtype=float)
        else:
            return transpose(stack((self._lo, self._hi)))

    @property
    def scalar(self):
        """Check if the interval is wide sense scalar

        note:
            wide sense: I(1,2) and I([1],[2]) are both scalars
        """
        return (self.shape == ()) | (self.shape == (1,))

    @property
    def is_scalar(self):
        """Check if the interval is a strict-sense scalar

        note:
            strict sense: I(1,2) is a scalar, but I([1],[2]) is not
        """
        return self.shape == ()

    @property
    def shape(self):
        return self.__shape

    @property
    def ndim(self):
        return len(self.__shape)

    # * -------------- ARITHMETIC -------------- *#
    # unary operators #
    def __neg__(self):
        return Interval(-self.hi, -self.lo)

    def __pos__(self):
        return self

    # binary operators #
    def __add__(self, other):
        otherType = other.__class__.__name__
        if (otherType == "ndarray") | (otherType in NUMERIC_TYPES):
            lo, hi = self.lo + other, self.hi + other
        elif otherType == "Interval":
            lo, hi = self.lo + other.lo, self.hi + other.hi
        else:
            return NotImplemented  # TypeError: unsupported operand type(s) for +: 'int' and 'Interval' (for example)
        return Interval(lo, hi)

    def __radd__(self, left):
        leftType = left.__class__.__name__
        if (leftType == "ndarray") | (leftType in NUMERIC_TYPES):
            return self.__add__(left)
        else:
            return NotImplemented  # TypeError: unsupported operand type(s) for +: 'int' and 'Interval' (for example)

    def __sub__(self, other):
        otherType = other.__class__.__name__
        if (otherType == "ndarray") | (otherType in NUMERIC_TYPES):
            lo, hi = self.lo - other, self.hi - other
        elif otherType == "Interval":
            lo, hi = self.lo - other.hi, self.hi - other.lo
        else:
            NotImplemented
        return Interval(lo, hi)

    def __rsub__(self, left):
        leftType = left.__class__.__name__
        if (leftType == "ndarray") | (leftType in NUMERIC_TYPES):
            lo, hi = left - self.hi, left - self.lo
        else:
            return NotImplemented  # print("Error: not among the allowed types.")
        return Interval(lo, hi)

    def __mul__(self, other):
        otherType = other.__class__.__name__
        if otherType in NUMERIC_TYPES:
            if other >= 0:
                lo, hi = self.lo * other, self.hi * other
            else:
                lo, hi = self.hi * other, self.lo * other
        elif otherType == "ndarray":  # check self and other have same shape
            lo, hi = numpy.empty(self._lo.shape), numpy.empty(self._lo.shape)
            if len(other.shape) == 0:
                self.__mul__(float(other))  # safety net for ndarrays with no shape
            other_positive = other >= 0
            other_negative = other_positive == False
            lo[other_positive] = self.lo[other_positive] * other[other_positive]
            hi[other_positive] = self.hi[other_positive] * other[other_positive]
            lo[other_negative] = self.hi[other_negative] * other[other_negative]
            hi[other_negative] = self.lo[other_negative] * other[other_negative]
        elif otherType == "Interval":
            lo, hi = multiply(self, other)
        else:
            return NotImplemented
        return Interval(lo, hi)

    def __rmul__(self, left):
        leftType = left.__class__.__name__
        if (leftType == "ndarray") | (leftType in NUMERIC_TYPES) | np.isscalar(left):
            return self.__mul__(left)
        else:
            return NotImplemented

    def __truediv__(self, other):
        otherType = other.__class__.__name__
        if otherType in NUMERIC_TYPES:
            if other == 0:
                raise ZeroDivisionError
            if other > 0:
                lo, hi = self.lo / other, self.hi / other
            else:
                lo, hi = self.hi / other, self.lo / other
        elif otherType == "ndarray":
            lo, hi = numpy.empty(self._lo.shape), numpy.empty(self._lo.shape)
            if any(other.flatten() == 0):
                raise ZeroDivisionError
            other_positive = other > 0
            other_negative = other_positive == False
            lo[other_positive] = self.lo[other_positive] / other[other_positive]
            hi[other_positive] = self.hi[other_positive] / other[other_positive]
            lo[other_negative] = self.hi[other_negative] / other[other_negative]
            hi[other_negative] = self.lo[other_negative] / other[other_negative]
            pass
        elif otherType == "Interval":
            lo, hi = divide(self, other)
        else:
            NotImplemented
        return Interval(lo, hi)

    def __rtruediv__(self, left):
        leftType = left.__class__.__name__
        # lo,hi = numpy.empty(self._lo.shape),numpy.empty(self._hi.shape)
        self_lo, self_hi = self.lo, self.hi
        self_straddle_zero = numpy.any(
            (self_lo.flatten() <= 0) & (self_hi.flatten() >= 0)
        )
        if self_straddle_zero:
            raise ZeroDivisionError
        if (leftType == "ndarray") | (leftType in NUMERIC_TYPES):
            if left >= 0:
                lo, hi = left / self_hi, left / self_lo
            else:
                lo, hi = left / self_lo, left / self_hi
        else:
            return NotImplemented
        return Interval(lo, hi)

    def __pow__(self, other):
        otherType = other.__class__.__name__
        if otherType in INTEGERS:
            a, b = numpy.asarray(self.lo**other), numpy.asarray(
                self.hi**other
            )  # a2,b2 = a**2, b**2
            if other % 2 == 0:  # even power
                lo = zeros(
                    a.shape
                )  # numpy.max([numpy.min([a,b],axis=0),numpy.zeros(a.shape)],axis=0)
                lo[self < 0] = b[self < 0]
                lo[self > 0] = a[self > 0]
                hi = numpy.max([a, b], axis=0)
            else:  # odd power
                lo = numpy.min([a, b], axis=0)
                hi = numpy.max([a, b], axis=0)
        else:
            raise NotImplementedError("Not implemented yet")
        return Interval(lo, hi)

    def __lt__(self, other):
        return hi(self) < lo(other)

    def __rlt__(self, left):
        return hi(left) < lo(self)

    def __gt__(self, other):
        return lo(self) > hi(other)

    def __rgt__(self, left):
        return lo(left) > hi(self)

    def __le__(self, other):
        return hi(self) <= lo(other)

    def __rle__(self, left):
        return hi(left) <= lo(self)

    def __ge__(self, other):
        return lo(self) >= hi(other)

    def __rge__(self, left):
        return lo(left) >= hi(self)

    def __eq__(self, other):
        return (lo(self) == lo(other)) & (hi(self) == hi(other))

    def __ne__(self, other):
        return not (self == other)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented
        if "out" in kwargs and kwargs["out"] is not None:
            return NotImplemented

        if ufunc is np.sin:
            return self.sin()
        if ufunc is np.cos:
            return self.cos()
        if ufunc is np.tan:
            return self.tan()
        if ufunc is np.exp:
            return self.exp()
        if ufunc is np.sqrt:
            return self.sqrt()
        if ufunc is np.log:
            return self.log()

        return NotImplemented

    # * -------------- unary functions -------------- *#

    def abs(self):
        from .methods import abs as iabs

        return iabs(self)

    def sqrt(self):
        from .methods import sqrt as isqrt

        return isqrt(self)

    def exp(self):
        from .methods import exp as iexp

        return iexp(self)

    def log(self):
        from .methods import log as ilog

        return ilog(self)

    def sin(self):
        from .methods import sin as isin

        return isin(self)

    def cos(self):
        from .methods import cos as icos

        return icos(self)

    def tan(self):
        from .methods import tan as itan

        return itan(self)

    @classmethod
    def from_meanform(cls, x, half_width):
        if np.isscalar(x):
            return cls(x - half_width, x + half_width)
        else:
            x = np.asarray(x)
            half_width = np.asarray(half_width)
            return cls(lo=x - half_width, hi=x + half_width)

    def save_json(
        self, filename: str, comment: str = None, save_dir: str | Path = "."
    ) -> None:
        """
        Save the interval object to a JSON5 file.

        Args:
            filename (str): The name of the file (without extension) to save the interval object to.
            comment (str, optional): A comment to include at the top of the file.
            save_dir (str | Path, optional): Directory where the file should be saved. Defaults to current directory.

        Note:
            The file is saved with a `.json5` extension.

        Example:
            >>> a.save_json("interval_data", comment="This is interval data", save_dir="results/")
        """
        # Ensure save_dir is a Path object and create directory if missing
        save_path = Path(save_dir)
        save_path.mkdir(parents=True, exist_ok=True)

        # Construct the full path with .json5 extension
        file_path = save_path / f"{filename}.json5"

        data = self.to_numpy().tolist()

        with open(file_path, "w", encoding="utf-8") as f:
            if comment:
                f.write(f"// {comment}\n")
            json5.dump(data, f, indent=2)


# * -------------- lightweight Interval


""" for more compatability with possibility of newer additions """
# class LightweightInterval(Interval):
#     def __init__(self, *args, **kwargs):
#         kwargs["do_heavy_checks"] = False
#         super().__init__(*args, **kwargs)

""" clearer right-now logic """


class LightweightInterval(Interval):
    def __init__(self, lo, hi=None):
        super().__init__(lo, hi, do_heavy_checks=False)

        # # Store the shape (this is just lo.shape, because they have the same shape)
        # self.__shape = ()

    def __repr__(self):
        return f"[{self._lo},{self._hi}]"

    @property
    def lo(self) -> Union[ndarray, float]:
        return self._lo

    # if len(self.shape)==0: return self._lo
    # return self._lo # return transpose(transpose(self.__val)[0]) # from shape (3,7,2) to (2,7,3) to (3,7)
    @property
    def hi(self) -> Union[ndarray, float]:
        return self._hi


# Properties or maybe attributes of the interval class. These apply to all interval-like objects.

#####################################################################################
# methods.py
#####################################################################################
# Interval to float methods, Unary.

# def iterator(x:Interval) -> Interval:
#     lo_iter,hi_iter = numpy.nditer(x.lo()),numpy.nditer(x.hi())
#     while True: yield Interval(lo=next(lo_iter),hi=next(hi_iter))


def is_Interval(x: Any) -> bool:
    x_class_name = x.__class__.__name__
    return x_class_name == "Interval"


def interval_degenerate(vec_interval):
    if vec_interval.is_degenerate():
        return vec_interval.lo
    return vec_interval


def lo(x: Interval) -> Union[float, ndarray]:
    """
    Return the left endpoint of an Interval object.

    If x is not of class Interval, input is returned.

    """
    if is_Interval(x):
        return x.lo
    return x


def hi(x: Interval) -> Union[float, ndarray]:
    """
    Return the right endpoint of an Interval object.

    If x is not of class Interval, input is returned.

    """
    if is_Interval(x):
        return x.hi
    return x


def width(x: Interval) -> Union[float, ndarray]:
    """
    Return the width of an Interval object.

    If x is not of class Interval, input is returned.

    """
    if is_Interval(x):
        return hi(x) - lo(x)
    return x


def rad(x: Interval) -> Union[float, ndarray]:
    """
    Return the radius of an Interval object.

    If x is not of class Interval, input is returned.

    """
    if is_Interval(x):
        return (hi(x) - lo(x)) / 2
    return x


def mid(x: Interval) -> Union[float, ndarray]:
    """
    Return the midpoint of an Interval.

    If x is not of class Interval, input is returned.

    """
    if is_Interval(x):
        return (hi(x) + lo(x)) / 2
    return x


def mig(x):
    return np.max(np.abs(x.lo), np.abs(x.hi))  # mignitude


def mag(x):
    return np.min(np.abs(x.lo), np.abs(x.hi))  # magnitude



from __future__ import annotations
from typing import TYPE_CHECKING
import warnings
import numpy as np
from abc import ABC, abstractmethod
from .params import Params
import matplotlib.pyplot as plt
from .intervals.number import Interval as I
from numbers import Number
import operator
from .utils import (
    condensation,
    find_nearest,
    get_mean_var_from_ecdf,
    is_increasing,
    left_right_switch,
    variance_bounds_via_lp,
    area_between_ecdfs,
)
import logging
from .context import get_current_dependency
from .mixins import NominalValueMixin
from contextlib import suppress

if TYPE_CHECKING:
    from pyuncertainnumber import Interval
    from .dss import DempsterShafer
    from .ecdf import eCDF_bundle
    from typing import Self

# Configure the logging system with a simple format
logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s: %(message)s",
)


def bound_steps_check(bound):
    # condensation needed
    if len(bound) > Params.steps:
        bound = condensation(bound, Params.steps)
    elif len(bound) < Params.steps:
        # 'next' kind interpolation needed
        from .constructors import interpolate_p

        p_lo, bound = interpolate_p(
            p=np.linspace(Params.p_lboundary, Params.p_hboundary, len(bound)), q=bound
        )
    return bound


# * --------------------- constructors ---------------------*#
def pbox_from_extredists(rvs, shape="beta", extre_bound_params=None):
    """transform into pbox object from extreme bounds parameterised by `sps.dist`

    args:
        rvs (list): list of scipy.stats.rv_continuous objects"""

    # x_sup
    bounds = [rv.ppf(Params.p_values) for rv in rvs]

    if bounds[0][-1] > bounds[1][-1]:
        # swap left and right bounds
        bounds[0], bounds[1] = bounds[1], bounds[0]

    if extre_bound_params is not None:
        print(extre_bound_params)

    return Staircase(
        left=bounds[0],
        right=bounds[1],
        shape=shape,
    )


def pbox_from_ecdf_bundle(lower_bound: eCDF_bundle, upper_bound: eCDF_bundle) -> Pbox:
    """Construct a p-box from two empirical CDF bundles as the extreme bounds"""
    return Staircase.from_CDFbundle(lower_bound, upper_bound)


def naive_frechet_pbox(x, y, op) -> Staircase:
    """A wrapper that returns a Pbox from the naive Frechet operation

    note:
        old implementation from pba.r
    """
    from .operation import naive_frechet_op

    Zu, Zd = naive_frechet_op(x, y, op)
    p = Staircase(left=Zu, right=Zd)
    return p


# deprecated
# def new_naive_frechet_pbox(x, y, op) -> Staircase:
#     """A wrapper that returns a Pbox from the naive Frechet operation"""
#     from .operation import new_naive_frechet_op

#     Zu, Zd = new_naive_frechet_op(x, y, op)
#     p = Staircase(left=Zu, right=Zd)
#     return p


def vectorised_naive_frechet_pbox(x, y, op) -> Staircase:
    """A wrapper that returns a Pbox from the naive Frechet operation"""
    from .operation import new_vectorised_naive_frechet_op

    Zu, Zd = new_vectorised_naive_frechet_op(x, y, op)
    p = Staircase(left=Zu, right=Zd)
    return p


def classic_frechet_pbox(x, y, op) -> Staircase:
    """this corresponds to the Frank, Nelson and Sklar Frechet bounds implementation"""
    from .operation import frechet_op

    left, right = frechet_op(x, y, op)
    p = Staircase(left=left, right=right)
    return p


def straddle_frechet_pbox(x, y):
    """bespoke Frechet for multiplcation when anyone straddles 0"""
    from .aggregation import imposition

    warnings.warn(
        "Multiplication of a pbox straddling zero needs attention",
        UserWarning,
    )
    naive_base_p = vectorised_naive_frechet_pbox(x, y, operator.mul)
    balch_p = x.balchprod(y)
    imp_p = imposition(naive_base_p, balch_p)
    return imp_p


def nagative_frechet_pbox(x, y):
    if x.hi <= 0 or y.hi <= 0:
        warnings.warn("negative pbox encountered")
        a = -x if x.hi <= 0 else x
        b = -y if y.hi <= 0 else y
        result = classic_frechet_pbox(a, b, operator.mul)
        return -result if (x.hi <= 0) ^ (y.hi <= 0) else result
    else:
        raise Exception("Not nagative Frechet case")


def frechet_pbox_mul(x, y):
    """the overall pbox"""

    if x.straddles_zero() or y.straddles_zero():  # if any one straddles
        if y.straddles_zero():  # y shall be straddle
            return straddle_frechet_pbox(x, y)
        else:
            return straddle_frechet_pbox(y, x)
    # elif x.is_zero() or y.is_zero():
    #     return 0
    elif x.hi <= 0 or y.hi <= 0:
        return nagative_frechet_pbox(x, y)
    else:  # both positive
        return classic_frechet_pbox(x, y, operator.mul)


class Pbox(NominalValueMixin, ABC):
    """a base class for Pbox

    danger:
        this is an abstract class and should not be instantiated directly.

        .. seealso::

            :class:`pbox_abc.Staircase` and :class:`pbox_abc.Leaf` for concrete implementations.
    """

    def __init__(
        self,
        left: np.ndarray | list,
        right: np.ndarray | list,
        steps=Params.steps,
        mean=None,
        var=None,
        p_values=None,
    ):
        left, right = left_right_switch(left, right)
        self.left = np.array(left)
        self.right = np.array(right)
        self.steps = steps
        self.mean = mean
        self.var = var
        # we force the steps but allow the p_values to be flexible
        self._pvalues = p_values if p_values is not None else Params.p_values
        self.post_init_check()

    # * --------------------- setup ---------------------*#

    @abstractmethod
    def _init_moments(self):
        pass

    def _init_range(self):
        self._range = I(min(self.left), max(self.right))

    def post_init_check(self):

        self.steps_check()

        if (not is_increasing(self.left)) or (not is_increasing(self.right)):
            raise Exception("Left and right arrays must be increasing")

        # pass along moments information
        if (self.mean is None) or (self.var is None):
            self._init_moments()

        self._init_range()

        self.degenerate_flag()

    def steps_check(self):

        assert len(self.left) == len(
            self.right
        ), "Length of lower and upper bounds is not consistent"

    def _compute_nominal_value(self):
        return np.round(self.mean.mid, 3)

    def degenerate_flag(self) -> bool:
        """check if the pbox is degenerate (i.e. left == right everywhere)"""
        self._degenerate = np.array_equal(self.left, self.right)

    @property
    def degenerate(self) -> bool:
        return self._degenerate

    @property
    def p_values(self):
        return self._pvalues

    @property
    def range(self):
        return self._range

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, value):
        self._left = bound_steps_check(value)
        self.steps = len(self._left)

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, value):
        self._right = bound_steps_check(value)
        self.steps = len(self._right)

    @property
    def lo(self):
        """Returns the left-most value in the interval"""
        return self.left[0]

    @property
    def hi(self):
        """Returns the right-most value in the interval"""
        return self.right[-1]

    @property
    def support(self):
        return self._range

    @property
    def median(self):
        return I(np.median(self.left), np.median(self.right))

    @property
    def enclosed_area(self):
        """the enclosed area between the two extreme cdfs"""
        return area_between_ecdfs(
            x_upper=self.left,
            p_upper=self._pvalues,
            x_lower=self.right,
            p_lower=self._pvalues,
        )

    # * --------------------- operators ---------------------*#

    def __iter__(self):
        return iter(self.to_interval())

    def __eq__(self, other):
        """Equality operator for Pbox objects

        note:
            - two pboxes are equal if their left and right bounds are equal
        """
        # equal = np.array_equal(self.left, other.left) and np.array_equal(
        #     self.right, other.right
        # )
        close = np.allclose(self.left, other.left) and np.allclose(
            self.right, other.right
        )
        return close

    def __contains__(self, item):
        if isinstance(item, Number):
            return (self.lo <= item) and (item <= self.hi)
        else:
            return (self.lo <= item.lo) and (item.hi <= self.hi)

    # * --------------------- functions ---------------------*#
    def to_interval(self):
        """discretise pbox into a vec-interval of length of default steps

        note:
            If desired a custom length of vec-interval as output, use `discretise()` method.
        """
        from .intervals.number import Interval as I

        return I(lo=self.left, hi=self.right)

    def to_dss(self, discretisation=Params.steps):
        """convert pbox to DempsterShafer object"""
        from .dss import DempsterShafer

        return DempsterShafer(
            self.to_interval(),
            np.repeat(a=(1 / discretisation), repeats=discretisation),
        )

    def to_numpy(self):
        """convert pbox to a 2D numpy array (n, 2) of left and right"""
        return np.stack((self.left, self.right), axis=1)


class Staircase(Pbox):
    """distribution free p-box"""

    def __init__(
        self,
        left,
        right,
        steps=200,
        mean=None,
        var=None,
        p_values=None,
    ):
        super().__init__(left, right, steps, mean, var, p_values)

    def _init_moments(self):
        """Initialize mean/var interval estimates.

        strategy:
            1) Try LP-based bounds.
            2) If that fails, try ECDF-based bounds.
            3) If that also fails, set to NaN intervals so the program continues.

        This function NEVER raises.
        """
        # Defaults
        mean_I = None
        var_I = None
        method_used = None
        errors = []

        # --- Attempt 1: LP bounds ---
        try:
            dict_moments = variance_bounds_via_lp(
                q_a=self.left,
                p_a=self._pvalues,
                q_b=self.right,
                p_b=self._pvalues,
                x_grid=np.linspace(self.lo, self.hi, 50),
            )
            self.mean_lo, self.mean_hi = dict_moments["mu_min"], dict_moments["mu_max"]
            self.var_lo, self.var_hi = dict_moments["var_min"], dict_moments["var_max"]
            mean_I = I(self.mean_lo, self.mean_hi)
            var_I = I(self.var_lo, self.var_hi)
            method_used = "lp_bounds"
        except Exception as e:
            errors.append(("lp_bounds", repr(e)))

        # --- Attempt 2: ECDF fallback (only if needed) ---
        if mean_I is None or var_I is None:
            try:
                mean_lo, var_lo = get_mean_var_from_ecdf(self.left, self._pvalues)
                mean_hi, var_hi = get_mean_var_from_ecdf(self.right, self._pvalues)
                self.mean_lo, self.mean_hi = mean_lo, mean_hi
                self.var_lo, self.var_hi = var_lo, var_hi
                mean_I = I(self.mean_lo, self.mean_hi)
                var_I = I(self.var_lo, self.var_hi)
                method_used = "ecdf_fallback"
            except Exception as e:
                errors.append(("ecdf_fallback", repr(e)))

        # --- Last resort: make it unambiguous and safe ---
        if mean_I is None or var_I is None:
            # Use NaN to signal “unknown/unavailable” without risking real-number collisions.
            mean_I = I(666, 666)
            var_I = I(666, 666)
            method_used = method_used or "unavailable"

        # --- Assign + annotate; nothing in here may raise ---
        self.mean = mean_I
        self.var = var_I
        # Optional: stash debug info for inspection; never let this crash.
        with suppress(Exception):
            self._moments_meta = {
                "method": method_used,
                "errors": errors,  # list of (stage, repr(error))
                "lo_hi": {
                    "mean": (
                        getattr(self, "mean_lo", None),
                        getattr(self, "mean_hi", None),
                    ),
                    "var": (
                        getattr(self, "var_lo", None),
                        getattr(self, "var_hi", None),
                    ),
                },
            }

    def __repr__(self):
        def format_interval(interval):
            try:
                return f"[{interval.lo:.3f}, {interval.hi:.3f}]"
            except Exception:
                return str(interval)

        mean_text = format_interval(self.mean)
        var_text = format_interval(self.var)
        range_text = format_interval(self._range)

        return f"Pbox ~ (range={range_text}, mean={mean_text}, var={var_text})"

    def plot(
        self,
        title=None,
        ax=None,
        style="box",
        fill_color="lightgray",
        bound_colors=None,
        bound_styles=None,  # e.g. ("--", ":")
        left_line_kwargs=None,  # e.g. {"linewidth": 2, "alpha": 0.9}
        right_line_kwargs=None,  # e.g. {"linewidth": 2, "alpha": 0.9}
        nuance="step",
        alpha=0.3,
        **kwargs,
    ):
        """default plotting function

        args:
            style (str): 'box' or 'simple'
            fill_color (str): color to fill the box (only for 'box' style)
            bound_colors (list): list of two colors for left and right bound lines
            bound_styles (list): list of two linestyles for left and right bound lines
            left_line_kwargs (dict): additional kwargs for left bound line
            right_line_kwargs (dict): additional kwargs for right bound line
            nuance (str): 'step' or 'curve' for bound line styles
            alpha (float): transparency level for the box fill (only for 'box' style)
            **kwargs: additional keyword arguments for the plot


        note:
            Two styles are supported: a 'box' with fill-in color and a 'simple' one without fill-in color.
            Color and linestyle of the bound lines can be customized via the `bound_styles`, `left_line_kwargs`, and `right_line_kwargs` parameters.
            The argument `nuance` controls whether the bound lines are plotted as step functions ('step') or smooth curves ('curve').


        example:
            >>> a = pba.normal([2, 6], [0.5, 1])
            >>> fig, ax = plt.subplots()
            >>> a.plot(ax=ax, style='simple')  # simple style without fill-in color
            >>> # box style with fill-in color and also customized bound colors
            >>> a.plot(ax=ax, style='box',
            ...     fill_color='lightblue',
            ...     bound_colors = ['lightblue', 'lightblue'],
            ...     bound_styles=("--", ":"),
            ...     alpha=0.5
            ... )
            >>> # customized left and right bound line styles
            >>> ax = pbox.plot(
            ...     left_line_kwargs={"linestyle": "--", "linewidth": 2},
            ...     right_line_kwargs={"linestyle": ":", "linewidth": 2, "alpha": 0.8},
            )

        """
        import matplotlib.pyplot as plt
        import matplotlib.patheffects as pe  # optional; for "shaded/halo" line effects
        from .utils import CustomEdgeRectHandler

        if ax is None:
            fig, ax = plt.subplots()

        p_axis = self._pvalues if self._pvalues is not None else Params.p_values
        plot_bound_colors = bound_colors if bound_colors is not None else ["g", "b"]

        # defaults
        if bound_styles is None:
            bound_styles = ("solid", "solid")
        left_line_kwargs = {} if left_line_kwargs is None else dict(left_line_kwargs)
        right_line_kwargs = {} if right_line_kwargs is None else dict(right_line_kwargs)

        # ensure color + linestyle are set unless user overrode them
        left_defaults = {"c": plot_bound_colors[0], "linestyle": bound_styles[0]}
        right_defaults = {"c": plot_bound_colors[1], "linestyle": bound_styles[1]}
        # user kwargs take precedence
        left_kwargs = {**left_defaults, **left_line_kwargs}
        right_kwargs = {**right_defaults, **right_line_kwargs}

        def display_box(nuance, label=None):
            """display two F curves plus the top-bottom horizontal lines"""

            if nuance == "step":
                step_kwargs_left = {"where": "post", **left_kwargs}
                step_kwargs_right = {"where": "post", **right_kwargs}
                if label is not None:
                    step_kwargs_left["label"] = label

                (line_left,) = ax.step(self.left, p_axis, **step_kwargs_left)
                (line_right,) = ax.step(self.right, p_axis, **step_kwargs_right)
            elif nuance == "curve":
                curve_kwargs_left = {**left_kwargs}
                curve_kwargs_right = {**right_kwargs}
                if label is not None:
                    curve_kwargs_left["label"] = label

                (line_left,) = ax.plot(self.left, p_axis, **curve_kwargs_left)
                (line_right,) = ax.plot(self.right, p_axis, **curve_kwargs_right)
            else:
                raise ValueError("nuance must be either 'step' or 'curve'")

            # horizontal caps (use right/left kwargs for consistent style/color)
            ax.plot([self.left[0], self.right[0]], [0, 0], **right_kwargs)
            ax.plot([self.left[-1], self.right[-1]], [1, 1], **left_kwargs)

            if label is not None:
                ax.legend(
                    handler_map={line_left: CustomEdgeRectHandler()}
                )  # regular use

        if title is not None:
            ax.set_title(title)

        if style == "box":
            ax.fill_betweenx(
                y=p_axis,
                x1=self.left,
                x2=self.right,
                interpolate=True,
                color=fill_color,
                alpha=alpha,
                **kwargs,
            )
            display_box(nuance, label=None)
            if "label" in kwargs:
                ax.legend(loc="best")
        elif style == "simple":
            display_box(nuance, label=kwargs.get("label"))
        else:
            raise ValueError("style must be either 'simple' or 'box'")

        ax.set_xlabel(r"$x$")
        ax.set_ylabel(r"$\Pr(X \leq x)$")
        return ax

    # backup old plot function
    # def plot(
    #     self,
    #     title=None,
    #     ax=None,
    #     style="box",
    #     fill_color="lightgray",
    #     bound_colors=None,
    #     nuance="step",
    #     alpha=0.3,
    #     **kwargs,
    # ):
    #     """default plotting function

    #     args:
    #         style (str): 'box' or 'simple'
    #     """
    #     from .utils import CustomEdgeRectHandler

    #     if ax is None:
    #         fig, ax = plt.subplots()

    #     p_axis = self._pvalues if self._pvalues is not None else Params.p_values
    #     plot_bound_colors = bound_colors if bound_colors is not None else ["g", "b"]

    #     def display_box(nuance, label=None):
    #         """display two F curves plus the top-bottom horizontal lines"""

    #         if nuance == "step":
    #             step_kwargs = {
    #                 "c": plot_bound_colors[0],
    #                 "where": "post",
    #             }

    #             if label is not None:
    #                 step_kwargs["label"] = label

    #             # Make the plot
    #             (line,) = ax.step(self.left, p_axis, **step_kwargs)
    #             ax.step(self.right, p_axis, c=plot_bound_colors[1], where="post")
    #             ax.plot([self.left[0], self.right[0]], [0, 0], c=plot_bound_colors[1])
    #             ax.plot([self.left[-1], self.right[-1]], [1, 1], c=plot_bound_colors[0])
    #         elif nuance == "curve":
    #             smooth_curve_kwargs = {
    #                 "c": plot_bound_colors[0],
    #             }

    #             if label is not None:
    #                 smooth_curve_kwargs["label"] = label

    #             (line,) = ax.plot(self.left, p_axis, **smooth_curve_kwargs)
    #             ax.plot(self.right, p_axis, c=plot_bound_colors[1])
    #             ax.plot([self.left[0], self.right[0]], [0, 0], c=plot_bound_colors[1])
    #             ax.plot([self.left[-1], self.right[-1]], [1, 1], c=plot_bound_colors[0])
    #         else:
    #             raise ValueError("nuance must be either 'step' or 'curve'")
    #         if label is not None:
    #             ax.legend(handler_map={line: CustomEdgeRectHandler()})  # regular use

    #     if title is not None:
    #         ax.set_title(title)
    #     if style == "box":
    #         ax.fill_betweenx(
    #             y=p_axis,
    #             x1=self.left,
    #             x2=self.right,
    #             interpolate=True,
    #             color=fill_color,
    #             alpha=alpha,
    #             **kwargs,
    #         )
    #         display_box(nuance, label=None)
    #         if "label" in kwargs:
    #             ax.legend(loc="best")
    #     elif style == "simple":
    #         display_box(nuance, label=kwargs["label"] if "label" in kwargs else None)
    #     else:
    #         raise ValueError("style must be either 'simple' or 'box'")
    #     ax.set_xlabel(r"$x$")
    #     ax.set_ylabel(r"$\Pr(X \leq x)$")
    #     return ax

    #### put something below ####
    def plot_reverse_axis(
        self,
        title=None,
        ax=None,
        style="box",
        fill_color="lightgray",
        bound_colors=None,
        nuance="step",
        alpha=0.3,
        orientation="xy",  # NEW: "xy" (default) or "yx" (swap axes)
        invert_xaxis=True,
        **kwargs,
    ):
        """A testing plotting function that can swap quantile and probability axes.

        args:
            style (str): 'box' or 'simple'
            orientation (str): 'xy' keeps x on horizontal and Pr(X<=x) on vertical;
                            'yx' swaps them.
        """
        from .utils import CustomEdgeRectHandler
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()

        p_axis = self._pvalues if self._pvalues is not None else Params.p_values
        plot_bound_colors = bound_colors if bound_colors is not None else ["g", "b"]

        def display_box(nuance, label=None):
            """display two F curves plus the top-bottom horizontal/vertical lines,
            depending on orientation.
            """
            if orientation == "xy":
                # x = left/right, y = p
                if nuance == "step":
                    step_kwargs = {"c": plot_bound_colors[0], "where": "post"}
                    if label is not None:
                        step_kwargs["label"] = label
                    (line,) = ax.step(self.left, p_axis, **step_kwargs)
                    ax.step(self.right, p_axis, c=plot_bound_colors[1], where="post")
                elif nuance == "curve":
                    curve_kwargs = {"c": plot_bound_colors[0]}
                    if label is not None:
                        curve_kwargs["label"] = label
                    (line,) = ax.plot(self.left, p_axis, **curve_kwargs)
                    ax.plot(self.right, p_axis, c=plot_bound_colors[1])
                else:
                    raise ValueError("nuance must be either 'step' or 'curve'")
                ax.plot([self.left[0], self.right[0]], [0, 0], c=plot_bound_colors[1])
                ax.plot([self.left[-1], self.right[-1]], [1, 1], c=plot_bound_colors[0])

            elif orientation == "yx":
                # x = p, y = left/right  (axes swapped)
                if nuance == "step":
                    step_kwargs = {"c": plot_bound_colors[0], "where": "post"}
                    if label is not None:
                        step_kwargs["label"] = label
                    (line,) = ax.step(p_axis, self.left, **step_kwargs)
                    ax.step(p_axis, self.right, c=plot_bound_colors[1], where="post")
                elif nuance == "curve":
                    curve_kwargs = {"c": plot_bound_colors[0]}
                    if label is not None:
                        curve_kwargs["label"] = label
                    (line,) = ax.plot(p_axis, self.left, **curve_kwargs)
                    ax.plot(p_axis, self.right, c=plot_bound_colors[1])
                else:
                    raise ValueError("nuance must be either 'step' or 'curve'")
                ax.plot([0, 0], [self.left[0], self.right[0]], c=plot_bound_colors[1])
                ax.plot([1, 1], [self.left[-1], self.right[-1]], c=plot_bound_colors[0])
            else:
                raise ValueError("orientation must be 'xy' or 'yx'")

            if label is not None:
                ax.legend(handler_map={line: CustomEdgeRectHandler()})

        if title is not None:
            ax.set_title(title)

        if style == "box":
            if orientation == "xy":
                ax.fill_betweenx(
                    y=p_axis,
                    x1=self.left,
                    x2=self.right,
                    interpolate=True,
                    color=fill_color,
                    alpha=alpha,
                    **kwargs,
                )
            else:  # 'yx'
                ax.fill_between(
                    x=p_axis,
                    y1=self.left,
                    y2=self.right,
                    interpolate=True,
                    color=fill_color,
                    alpha=alpha,
                    **kwargs,
                )
            display_box(nuance, label=None)
            if "label" in kwargs:
                ax.legend(loc="best")
        elif style == "simple":
            display_box(nuance, label=kwargs["label"] if "label" in kwargs else None)
        else:
            raise ValueError("style must be either 'simple' or 'box'")

        if orientation == "xy":
            ax.set_xlabel(r"$x$")
            ax.set_ylabel(r"$\Pr(X \leq x)$")
        else:
            ax.set_xlabel(r"$\Pr(X \leq x)$")
            ax.set_ylabel(r"$x$")
            if invert_xaxis:
                ax.invert_xaxis()  # NEW LINE — reverses new x-axis (1 → 0)
            ax.yaxis.tick_right()
            ax.yaxis.set_label_position("right")

        return ax

    ### above ###

    def plot_outside_legend(
        self,
        title=None,
        ax=None,
        style="box",
        fill_color="lightgray",
        bound_colors=None,
        nuance="step",
        alpha=0.3,
        **kwargs,
    ):
        """a specific variant of `plot()` which is used for scipy proceeding only.

        args:
            style (str): 'box' or 'simple'
        """
        from .utils import CustomEdgeRectHandler

        if ax is None:
            fig, ax = plt.subplots()

        p_axis = self._pvalues if self._pvalues is not None else Params.p_values
        plot_bound_colors = bound_colors if bound_colors is not None else ["g", "b"]

        def display_box(nuance, label=None):
            """display two F curves plus the top-bottom horizontal lines"""

            if nuance == "step":
                step_kwargs = {
                    "c": plot_bound_colors[0],
                    "where": "post",
                }

                if label is not None:
                    step_kwargs["label"] = label

                # Make the plot
                (line,) = ax.step(self.left, p_axis, **step_kwargs)
                ax.step(self.right, p_axis, c=plot_bound_colors[1], where="post")
                ax.plot([self.left[0], self.right[0]], [0, 0], c=plot_bound_colors[1])
                ax.plot([self.left[-1], self.right[-1]], [1, 1], c=plot_bound_colors[0])
            elif nuance == "curve":
                smooth_curve_kwargs = {
                    "c": plot_bound_colors[0],
                }

                if label is not None:
                    smooth_curve_kwargs["label"] = label

                (line,) = ax.plot(self.left, p_axis, **smooth_curve_kwargs)
                ax.plot(self.right, p_axis, c=plot_bound_colors[1])
                ax.plot([self.left[0], self.right[0]], [0, 0], c=plot_bound_colors[1])
                ax.plot([self.left[-1], self.right[-1]], [1, 1], c=plot_bound_colors[0])
            else:
                raise ValueError("nuance must be either 'step' or 'curve'")
            if label is not None:
                # ax.legend(handler_map={line: CustomEdgeRectHandler()})  # regular use
                # Put a legend to the right of the current axis
                ax.legend(
                    handler_map={line: CustomEdgeRectHandler()},
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                )  # onetime use

        if title is not None:
            ax.set_title(title)
        if style == "box":
            ax.fill_betweenx(
                y=p_axis,
                x1=self.left,
                x2=self.right,
                interpolate=True,
                color=fill_color,
                alpha=alpha,
                **kwargs,
            )
            display_box(nuance, label=None)
            if "label" in kwargs:
                ax.legend(loc="best")
        elif style == "simple":
            display_box(nuance, label=kwargs["label"] if "label" in kwargs else None)
        else:
            raise ValueError("style must be either 'simple' or 'box'")
        ax.set_xlabel(r"$x$")
        ax.set_ylabel(r"$\Pr(X \leq x)$")
        return ax

    def display(self, *args, **kwargs):
        self.plot(*args, **kwargs)
        plt.show()

    def plot_probability_bound(self, x: float, ax=None, **kwargs):
        """plot the probability bound at a certain quantile x

        note:
            - a vertical line
        """

        if ax is None:
            fig, ax = plt.subplots()

        p_lo = self.cdf(x).lo
        p_hi = self.cdf(x).hi
        self.plot(ax=ax, **kwargs)

        ax.plot(
            [x, x],
            [p_lo, p_hi],
            c="r",
            label="probability bound",
            zorder=50,
        )
        ax.scatter(x, p_lo, c="r", marker="^", zorder=50)
        ax.scatter(x, p_hi, c="r", marker="v", zorder=50)
        return ax

    def plot_quantile_bound(self, p: float, ax=None, **kwargs):
        """plot the quantile bound at a certain probability level p

        note:
            - a horizontal line
        """
        if ax is None:
            fig, ax = plt.subplots()

        x_lo, x_hi = self.alpha_cut(p).lo, self.alpha_cut(p).hi

        self.plot(ax=ax, **kwargs)

        ax.plot(
            [x_lo, x_hi],
            [p, p],
            c="r",
            label="probability bound",
            zorder=50,
        )
        ax.scatter(x_lo, p, c="r", marker=">", zorder=50)
        ax.scatter(x_hi, p, c="r", marker="<", zorder=50)
        return ax

    # * --------------------- constructors ---------------------*#
    @classmethod
    def from_CDFbundle(cls, a, b):
        """pbox from two emipirical CDF bundle

        args:
            - a : CDF bundle of lower extreme F;
            - b : CDF bundle of upper extreme F;
        """
        from .constructors import interpolate_p
        from .utils import extend_ecdf

        a = extend_ecdf(a)
        b = extend_ecdf(b)

        p_lo, q_lo = interpolate_p(a.probabilities, a.quantiles)
        p_hi, q_hi = interpolate_p(b.probabilities, b.quantiles)
        return cls(left=q_lo, right=q_hi, p_values=p_lo)

    # * --------------------- operators ---------------------*#

    def __neg__(self):
        return Staircase(
            left=sorted(-np.flip(self.right)),
            right=sorted(-np.flip(self.left)),
            mean=-self.mean,
            var=self.var,
        )

    def __add__(self, other):
        return self.add(other, dependency=get_current_dependency())

    def __radd__(self, other):
        return self.add(other, dependency=get_current_dependency())

    def __sub__(self, other):
        return self.sub(other, dependency=get_current_dependency())

    def __rsub__(self, other):
        return (-self).add(other, dependency=get_current_dependency())

    def __mul__(self, other):
        return self.mul(other, dependency=get_current_dependency())

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):

        return self.div(other, dependency=get_current_dependency())

    def __rtruediv__(self, other):
        try:
            return other * self.reciprocal()
        except:
            return NotImplemented

    def __pow__(self, other):
        return self.pow(other, dependency=get_current_dependency())

    def __rpow__(self, other: Number):
        """Power operation with the base as `other` and self as the exponent"""
        from functools import partial

        bar = partial(np.power, other)  # other as the base
        return self._unary_template(bar)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented
        if len(inputs) != 1 or inputs[0] is not self:
            return NotImplemented
        if "out" in kwargs and kwargs["out"] is not None:
            return NotImplemented

        if ufunc is np.sin:
            return self.sin()
        if ufunc is np.cos:
            return self.cos()
        if ufunc is np.tanh:
            return self.tanh()
        if ufunc is np.exp:
            return self.exp()
        if ufunc is np.sqrt:
            return self.sqrt()
        if ufunc is np.log:
            return self.log()
        if ufunc is np.reciprocal:
            return self.reciprocal()

        return NotImplemented

    # * --------------------- methods ---------------------*#

    def cdf(self, x: np.ndarray):
        """get the bounds on the cdf w.r.t x value

        args:
            x (array-like): x values
        """
        lo_ind = find_nearest(self.right, x)
        hi_ind = find_nearest(self.left, x)
        return I(lo=Params.p_values[lo_ind], hi=Params.p_values[hi_ind])

    def alpha_cut(self, alpha=0.5):
        """test the lightweight `alpha_cut` method

        args:
            alpha (array-like): probability levels
        """
        from .intervals.number import LightweightInterval as lwI
        from .intervals.number import Interval as I

        ind = find_nearest(Params.p_values, alpha)
        return I(lo=self.left[ind], hi=self.right[ind])

    def sample(self, n_sam):
        """LHS sampling by default"""
        from scipy.stats import qmc

        alpha = np.squeeze(qmc.LatinHypercube(d=1).random(n=n_sam))
        return self.alpha_cut(alpha)

    def discretise(self, n=None) -> Interval:
        """alpha-cut discretisation of the p-box without outward rounding

        args:
            n (int): number of steps to be used in the discretisation.

        return:
            vector Interval
        """

        if (n is None) or (n == Params.steps):
            return I(lo=self.left, hi=self.right)
        else:
            p_values = np.linspace(Params.p_lboundary, Params.p_hboundary, n)
            return self.alpha_cut(p_values)

    def outer_discretisation(self, n=None):
        """discretisation of a p-box to get intervals based on the scheme of outer approximation

        args:
            n (int): number of steps to be used in the discretisation

        note:
            `the_interval_list` will have length one less than that of default `p_values` (i.e. 100 and 99)

        return:
            the outer intervals in vec-Interval form
        """

        from .intervals.number import Interval as I
        from .intervals.number import LightweightInterval as lwI

        if n is not None:
            p_values = np.linspace(Params.p_lboundary, Params.p_hboundary, n)
        else:
            p_values = self._pvalues

        p_leftend = p_values[0:-1]
        p_rightend = p_values[1:]

        q_l = self.alpha_cut(p_leftend).left
        q_r = self.alpha_cut(p_rightend).right
        interval_vec = lwI(lo=q_l, hi=q_r)

        return interval_vec

    def condensation(self, n) -> Self:
        """ourter condensation of the pbox to reduce the number of steps and get a sparser staircase pbox

        args:
            n (int): number of steps to be used in the discretisation

        note:
            Have not thought about a better name so we call it `condensation` for now. Candidate names include 'approximation'.
            It will ouput a p-box and keep steps as 200 for computational consistency.

        example:
            >>> p.condensation(n=5)

        return:
            a staircase p-box that looks sparser but has the same number of steps
        """
        from .aggregation import stacking

        itvls = self.outer_discretisation(n)
        return stacking(itvls)

    def condense(self, n) -> DempsterShafer:
        """Another condensation function which has steps of n

        Compared to the above `condensation` method that ouputs a p-box and  keeps steps as 200 for computational consistency.
        This one condenses in a more literal manner, as in having n steps in the resulting Dempster-Shafer structure.
        """
        from .intervals.number import Interval as I
        from .dss import DempsterShafer as DSS

        condensed_x = self.condensation(n)

        # condensed bounds
        con_bd_l = np.unique(condensed_x.left)
        con_bd_r = np.unique(condensed_x.right)

        real_condensed_dss_x = DSS(
            intervals=I(
                lo=con_bd_l,
                hi=con_bd_r,
            ),
            masses=[1 / len(con_bd_l)] * len(con_bd_l),
        )
        return real_condensed_dss_x

    def truncate(self, a, b):
        """Truncate the Pbox to the range [a, b].

        example:
        >>> from pyuncertainnumber import pba
        >>> p = pba.normal([4, 9], 1)
        >>> tr = p.truncate(3, 8)
        >>> fig, ax = plt.subplots()
        >>> p.plot(ax=ax)
        >>> tr.plot(ax=ax, fill_color='r')
        >>> plt.show()
        """

        from .aggregation import _imposition
        from .pbox_free import min_max

        i = min_max(a, b, return_construct=True)
        return _imposition(self, i)

    def min(self, other, method="f"):
        """Returns a new Pbox object that represents the element-wise minimum of two Pboxes.

        args:
            - other: Another Pbox object or a numeric value.
            - method: Calculation method to determine the minimum. Can be one of 'f', 'p', 'o', 'i'.

        returns:
            Pbox
        """

        other = convert_pbox(other)
        match method:
            case "f":
                nleft = np.empty(self.steps)
                nright = np.empty(self.steps)
                for i in range(0, self.steps):
                    j = np.array(range(i, self.steps))
                    k = np.array(range(self.steps - 1, i - 1, -1))
                    nright[i] = min(list(self.right[j]) + list(other.right[k]))
                    jj = np.array(range(0, i + 1))
                    kk = np.array(range(i, -1, -1))
                    nleft[i] = min(list(self.left[jj]) + list(other.left[kk]))
            case "p":
                nleft = np.minimum(self.left, other.left)
                nright = np.minimum(self.right, other.right)
            case "o":
                nleft = np.minimum(self.left, np.flip(other.left))
                nright = np.minimum(self.right, np.flip(other.right))
            case "i":
                nleft = []
                nright = []
                for i in self.left:
                    for j in other.left:
                        nleft.append(np.minimum(i, j))
                for ii in self.right:
                    for jj in other.right:
                        nright.append(np.minimum(ii, jj))
        nleft.sort()
        nright.sort()

        return Staircase(left=nleft, right=nright)

    def max(self, other, method="f"):

        other = convert_pbox(other)
        match method:
            case "f":
                nleft = np.empty(self.steps)
                nright = np.empty(self.steps)
                for i in range(0, self.steps):
                    j = np.array(range(i, self.steps))
                    k = np.array(range(self.steps - 1, i - 1, -1))
                    nright[i] = max(list(self.right[j]) + list(other.right[k]))
                    jj = np.array(range(0, i + 1))
                    kk = np.array(range(i, -1, -1))
                    nleft[i] = max(list(self.left[jj]) + list(other.left[kk]))
            case "p":
                nleft = np.maximum(self.left, other.left)
                nright = np.maximum(self.right, other.right)
            case "o":
                nleft = np.maximum(self.left, np.flip(other.right))
                nright = np.maximum(self.right, np.flip(other.left))
            case "i":
                nleft = []
                nright = []
                for i in self.left:
                    for j in other.left:
                        nleft.append(np.maximum(i, j))
                for ii in self.right:
                    for jj in other.right:
                        nright.append(np.maximum(ii, jj))

        nleft.sort()
        nright.sort()

        return Staircase(left=nleft, right=nright)

    def get_PI(self, alpha: Number = 0.95, style="narrowest") -> Interval:
        """Compute the predictive interval at the coverage level of `alpha`

        args:
            alpha (Number): coverage level for the predictive interval, default is 0.95
            style (str): 'narrowest' or 'widest', default is 'narrowest'

        note:
            by default, narrowest predictive interval is returned;
            when the narrowest does not exist, a warning will the generated and then the widest is returned instead.

        example:
            >>> from pyuncertainnumber import pba
            >>> p = pba.normal([10, 15, 1])
            >>> p.get_PI(alpha=0.95, style='narrowest')
        """
        from pyuncertainnumber import Interval

        lo_cut_level = (1 - alpha) / 2  # 0.025
        hi_cut_level = 1 - lo_cut_level  # 0.975

        if style == "narrowest":
            hi = self.alpha_cut(hi_cut_level).lo
            lo = self.alpha_cut(lo_cut_level).hi
            try:
                return Interval(lo=lo, hi=hi)
            except Exception:
                logging.warning(
                    "narrowest predictive interval does not exist. Return 'widest' style instead."
                )
                # do the style == 'widest'
                hi = self.alpha_cut(hi_cut_level).hi
                lo = self.alpha_cut(lo_cut_level).lo
                return Interval(lo=lo, hi=hi)
        elif style == "widest":
            hi = self.alpha_cut(hi_cut_level).hi
            lo = self.alpha_cut(lo_cut_level).lo
            return Interval(lo=lo, hi=hi)

    # * --------------------- states --------------------- *#

    def straddles(self, N, endpoints=True) -> bool:
        """Check whether the p-box straddles a number N

        args:
            N (float): the Number to check
            endpoints (Boolean): Whether to include the endpoints within the check

        return:
            True
                If :math:`\\mathrm{left} \\leq N \\leq \mathrm{right}` (Assuming `endpoints=True`)
            False
                Otherwise

        note:
            This could affect the results of Frechet bounds
        """
        if endpoints:
            if min(self.left) <= N and max(self.right) >= N:
                return True
        else:
            if min(self.left) < N and max(self.right) > N:
                return True

        return False

    def straddles_zero(self) -> bool:
        """Checks specifically whether :math:`0` is within the p-box"""
        return self.straddles(0, False)

    def is_zero(self):
        return self.lo == 0

    def is_nagative(self):
        return self.hi <= 0

    # * --------------------- aggregations--------------------- *#
    def env(self, other):
        """computes the envelope of two Pboxes.

        args:
            other (Pbox)

        returns:
            - Pbox
        """

        nleft = np.minimum(self.left, other.left)
        nright = np.maximum(self.right, other.right)
        return Staircase(left=nleft, right=nright, steps=self.steps)

    def imp(self, other):
        """Returns the imposition of self with other pbox

        note:
            - binary imposition between two pboxes only
        """
        u = []
        d = []
        for sL, sR, oL, oR in zip(self.left, self.right, other.left, other.right):
            if max(sL, oL) > min(sR, oR):
                raise Exception(
                    "Imposition does not exist as high left greater than low right"
                )
            u.append(max(sL, oL))
            d.append(min(sR, oR))
        return Staircase(left=u, right=d)

    # * ---------------------unary operations--------------------- *#

    def _unary_template(self, f):
        l, r = f(self.left), f(self.right)
        return Staircase(left=l, right=r)

    def exp(self):
        """exponential function: e^x"""
        return self._unary_template(np.exp)

    def sqrt(self):
        """square root function: √x"""
        return self._unary_template(np.sqrt)

    def reciprocal(self):
        """Calculate the reciprocal of the pbox

        note:
            the pbox should not straddle zero, otherwise a warning is raised
        """

        if self.straddles_zero():
            warnings.warn(
                "Division of a pbox straddling zero needs attention", UserWarning
            )
        return Staircase(left=1 / np.flip(self.right), right=1 / np.flip(self.left))

    def log(self):
        """natural logarithm of the pbox

        note:
            - the pbox must be positive
        """
        if self.lo <= 0:
            raise ValueError("Logarithm is not defined for non-positive values")
        return self._unary_template(np.log)

    def sin(self):
        from .intervals.methods import sin

        itvls = sin(self.to_interval())
        return simple_stacking(itvls)

    def cos(self):
        from .intervals.methods import cos

        itvls = cos(self.to_interval())
        return simple_stacking(itvls)

    def tanh(self):
        from .intervals.methods import tanh

        itvls = tanh(self.to_interval())
        return simple_stacking(itvls)

    # * ---------------------binary operations--------------------- *#

    def add(self, other, dependency="f"):
        from .operation import (
            frechet_op,
            independent_op,
            perfect_op,
            opposite_op,
        )

        if isinstance(other, Number):
            return pbox_number_ops(self, other, operator.add)
        if is_un(other):
            other = convert_pbox(other)
        match dependency:
            case "f":
                nleft, nright = frechet_op(self, other, operator.add)
            case "p":
                nleft, nright = perfect_op(self, other, operator.add)
            case "o":
                nleft, nright = opposite_op(self, other, operator.add)
            case "i":
                nleft, nright = independent_op(self, other, operator.add)
            case _:
                raise ValueError(f"Unknown dependency type: {dependency!r}")
        nleft.sort()
        nright.sort()
        return Staircase(left=nleft, right=nright)

    def sub(self, other, dependency="f"):

        if dependency == "o":
            dependency = "p"
        elif dependency == "p":
            dependency = "o"

        return self.add(-other, dependency)

    def mul(self, other, dependency="f"):
        """Multiplication of uncertain numbers with the defined dependency dependency"""
        from .operation import (
            independent_op,
            perfect_op,
            opposite_op,
        )

        if isinstance(other, Number):
            return pbox_number_ops(self, other, operator.mul)
        if is_un(other):
            other = convert_pbox(other)

        match dependency:
            case "f":
                return frechet_pbox_mul(self, other)
            case "p":
                nleft, nright = perfect_op(self, other, operator.mul)
            case "o":
                nleft, nright = opposite_op(self, other, operator.mul)
            case "i":
                nleft, nright = independent_op(self, other, operator.mul)
        return Staircase(left=nleft, right=nright)

    def div(self, other, dependency="f"):

        if self.straddles_zero():
            warnings.warn(
                "Division of a pbox straddling zero needs attention", UserWarning
            )

        if dependency == "o":
            dependency = "p"
        elif dependency == "p":
            dependency = "o"

        return self.mul(1 / other, dependency)

    def pow(self, other, dependency="f"):
        """Exponentiation of uncertain numbers with the defined dependency dependency

        This suggests that the exponent (i.e. `other`) can also be an uncertain number.
        """
        from .operation import frechet_op, vectorized_cartesian_op

        if isinstance(other, Number):
            if self.straddles_zero():
                from pyuncertainnumber import pba

                itvls = self.to_interval()
                response = itvls**other
                return pba.stacking(response)
            else:
                return pbox_number_ops(self, other, operator.pow)

        if is_un(other):
            other = convert_pbox(other)

        match dependency:
            case "f":
                nleft, nright = frechet_op(self, other, operator.pow)
            case "p":
                nleft = self.left**other.left
                nright = self.right**other.right
            case "o":
                nleft = self.left ** np.flip(other.right)
                nright = self.right ** np.flip(other.left)
            case "i":
                nleft = vectorized_cartesian_op(self.left, other.left, operator.pow)
                nright = vectorized_cartesian_op(self.right, other.right, operator.pow)
        nleft.sort()
        nright.sort()
        return Staircase(left=nleft, right=nright)

    # TODO: add a function which can be imported in operation.py
    def balchprod(self, other):
        """Frechet convolution of two pboxes when any of them straddles zero"""

        if self.straddles_zero() and other.straddles_zero():
            x0 = self.lo
            y0 = other.lo
            xx0 = self - x0
            yy0 = other - y0
            a = frechet_pbox_mul(xx0, yy0)
            b1, b2 = y0 * xx0, x0 * yy0
            b = classic_frechet_pbox(b1, b2, operator.add)
            return classic_frechet_pbox(a, b, operator.add) + x0 * y0
        if self.straddles_zero():
            x0 = self.lo
            xx0 = self - x0
            a = frechet_pbox_mul(xx0, other)
            b = x0 * other
            return frechet_pbox_mul(a, b)
        if other.straddles_zero():
            y0 = other.lo
            yy0 = other - y0
            a = frechet_pbox_mul(self, yy0)
            b = self * y0
            return classic_frechet_pbox(a, b, operator.add)
        return frechet_pbox_mul(self, other)


class Leaf(Staircase):
    """parametric pbox"""

    def __init__(
        self,
        left=None,
        right=None,
        steps=200,
        mean=None,
        var=None,
        dist_params=None,
        shape=None,
    ):
        super().__init__(left, right, steps, mean, var)
        self.shape = shape
        self.dist_params = dist_params

    def _init_moments_range(self):
        print("not decided yet")

    def __repr__(self):
        base_repr = super().__repr__().rstrip(")")  # remove trailing ')'
        return f"{base_repr}, shape={self.shape})"  # added back trailing ')'

    def sample(self, n_sam):
        """sample from a parametric pbox or distribution"""

        s_i = super().sample(n_sam)
        if np.all(s_i.lo == s_i.hi):
            logging.info("samples generated from a precise distribution")
            return s_i.lo
        else:
            return s_i


class Cbox(Pbox):
    def __init__(self, left, right, steps=200):
        super().__init__(left, right, steps)


# * --------------------- module functions ---------------------*#


def is_un(un):
    """if the `un` is modelled by accepted constructs"""

    from .intervals.number import Interval
    from .dss import DempsterShafer
    from .distributions import Distribution

    return isinstance(un, Pbox | Interval | DempsterShafer | Distribution)


def convert_pbox(un):
    """transform the input un into a Pbox object

    note:
        - theorically 'un' can be {Interval, DempsterShafer, Distribution, float, int}
    """

    from .pbox_abc import Pbox
    from .dss import DempsterShafer
    from .distributions import Distribution
    from .intervals.number import Interval as I

    if isinstance(un, Pbox):
        return un
    elif isinstance(un, I):
        return un.to_pbox()
        # return Staircase(
        #     left=np.repeat(un.lo, Params.steps),
        #     right=np.repeat(un.hi, Params.steps),
        #     mean=un,
        #     var=I(0, (un.hi - un.lo) * (un.hi - un.lo) / 4),
        # )
    elif isinstance(un, Pbox):
        return un
    elif isinstance(un, Distribution):
        return un.to_pbox()
    elif isinstance(un, DempsterShafer):
        return un.to_pbox()
    else:
        raise TypeError(f"Unable to convert {type(un)} object to Pbox")


def pbox_number_ops(pbox: Pbox, n: Number, f: callable):
    # TODO: ask Scott. pbox sqrt operaton how to do?
    """blueprint for arithmetic between pbox and real numbers"""
    l = f(pbox.left, n)
    r = f(pbox.right, n)
    l = sorted(l)
    r = sorted(r)
    try:
        new_mean = f(pbox.mean, n)
    except:
        new_mean = None
    return Staircase(left=l, right=r, var=pbox.var)

    # Staircase(left=pbox.left + n, right=pbox.right + n)


def truncate(pbox, min, max):
    return pbox.truncate(min, max)


# * --------------------- unary functions ---------------------*#
def sin():
    pass


def cos():
    pass


def tanh():
    pass


def exp():
    pass


def log():
    pass


def sqrt():
    pass


# * --------------------- utility functions tmp ---------------------*#
def simple_stacking(itvls):
    """simple version of stacking vector Interval objects into pbox

    args:
        itvls (Interval): a vector Interval object to be stacked

    note:
        - only meant for quick use during development
        - see `stacking` function for production use
    """
    from .ecdf import get_ecdf, eCDF_bundle

    q1, p1 = get_ecdf(itvls.lo)
    q2, p2 = get_ecdf(itvls.hi)

    cdf1 = eCDF_bundle(q1, p1)
    cdf2 = eCDF_bundle(q2, p2)
    return Staircase.from_CDFbundle(cdf1, cdf2)


def inspect_pbox(pbox):
    """quickly inspect a pbox object"""
    print(pbox)
    pbox.display(nuance="curve")
